# type: ignore
#----------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela (tpoikela)
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#----------------------------------------------------------------------

from .uvm_tlm_imps import (UVM_BLOCKING_GET_IMP, UVM_BLOCKING_GET_PEEK_IMP, UVM_BLOCKING_PEEK_IMP,
   UVM_BLOCKING_PUT_IMP, UVM_BLOCKING_TRANSPORT_IMP, UVM_GET_IMP,
   UVM_GET_PEEK_IMP, UVM_NONBLOCKING_GET_IMP, UVM_NONBLOCKING_GET_PEEK_IMP,
   UVM_NONBLOCKING_PEEK_IMP, UVM_NONBLOCKING_PUT_IMP,
   UVM_NONBLOCKING_TRANSPORT_IMP, UVM_PEEK_IMP, UVM_PORT_COMMON,
   UVM_PUT_IMP, UVM_TRANSPORT_IMP)

from ..macros.uvm_tlm_defines import (UVM_TLM_BLOCKING_GET_MASK, UVM_TLM_BLOCKING_GET_PEEK_MASK,
    UVM_TLM_BLOCKING_MASTER_MASK, UVM_TLM_BLOCKING_PEEK_MASK,
    UVM_TLM_BLOCKING_PUT_MASK, UVM_TLM_BLOCKING_SLAVE_MASK,
    UVM_TLM_BLOCKING_TRANSPORT_MASK, UVM_TLM_GET_MASK,
    UVM_TLM_GET_PEEK_MASK, UVM_TLM_MASTER_MASK,
    UVM_TLM_NONBLOCKING_GET_MASK,
    UVM_TLM_NONBLOCKING_GET_PEEK_MASK,
    UVM_TLM_NONBLOCKING_MASTER_MASK,
    UVM_TLM_NONBLOCKING_PEEK_MASK, UVM_TLM_NONBLOCKING_PUT_MASK,
    UVM_TLM_NONBLOCKING_SLAVE_MASK,
    UVM_TLM_NONBLOCKING_TRANSPORT_MASK, UVM_TLM_PEEK_MASK,
    UVM_TLM_PUT_MASK, UVM_TLM_SLAVE_MASK, UVM_TLM_TRANSPORT_MASK)

#------------------------------------------------------------------------------
# Title: TLM Port Classes
#------------------------------------------------------------------------------
# The following classes define the TLM port classes.
#------------------------------------------------------------------------------


#------------------------------------------------------------------------------
#
# Class: uvm_*_port #(T)
#
# These unidirectional ports are instantiated by components that ~require~,
# or ~use~, the associated interface to convey transactions. A port can
# be connected to any compatible port, export, or imp port. Unless its
# ~min_size~ is 0, a port ~must~ be connected to at least one implementation
# of its associated interface.
#
# The asterisk in ~uvm_*_port~ is any of the following
#
#|  blocking_put
#|  nonblocking_put
#|  put
#|
#|  blocking_get
#|  nonblocking_get
#|  get
#|
#|  blocking_peek
#|  nonblocking_peek
#|  peek
#|
#|  blocking_get_peek
#|  nonblocking_get_peek
#|  get_peek
#
# Type parameters
#
# T - The type of transaction to be communicated by the export. The type T is not restricted
# to class handles and may be a value type such as int,enum,struct or similar.
#
# Ports are connected to interface implementations directly via
# <uvm_*_imp #(T,IMP)> ports or indirectly via hierarchical connections
# to <uvm_*_port #(T)> and <uvm_*_export #(T)> ports.
#
#------------------------------------------------------------------------------

# Function: new
#
# The ~name~ and ~parent~ are the standard <uvm_component> constructor arguments.
# The ~min_size~ and ~max_size~ specify the minimum and maximum number of
# interfaces that must have been connected to this port by the end of elaboration.
#
#|  function new (string name,
#|                uvm_component parent,
#|                int min_size=1,
#|                int max_size=1)

class UVMBlockingPutPort():
    # extends uvm_port_base #(uvm_tlm_if_base #(T,T));
    pass
UVMBlockingPutPort = UVM_PORT_COMMON(UVMBlockingPutPort, UVM_TLM_BLOCKING_PUT_MASK, "uvm_blocking_put_port")
UVM_BLOCKING_PUT_IMP('m_if', UVMBlockingPutPort)
#endclass

#class uvm_nonblocking_put_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_NONBLOCKING_PUT_MASK,"uvm_nonblocking_put_port")
#  `UVM_NONBLOCKING_PUT_IMP(this.m_if, T, t)
#endclass
class UVMNonBlockingPutPort():
    pass
UVMNonBlockingPutPort = UVM_PORT_COMMON(UVMNonBlockingPutPort, UVM_TLM_NONBLOCKING_PUT_MASK, "uvm_nonblocking_put_port")
UVM_NONBLOCKING_PUT_IMP('m_if', UVMNonBlockingPutPort)

#class uvm_put_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_PUT_MASK,"uvm_put_port")
#  `UVM_PUT_IMP(this.m_if, T, t)
#endclass
class UVMPutPort():
    pass
UVMPutPort = UVM_PORT_COMMON(UVMPutPort, UVM_TLM_PUT_MASK,"uvm_put_port")
UVM_PUT_IMP('m_if', UVMPutPort)

#class uvm_blocking_get_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_BLOCKING_GET_MASK,"uvm_blocking_get_port")
#  `UVM_BLOCKING_GET_IMP(this.m_if, T, t)
#endclass
class UVMBlockingGetPort():
    pass
UVMBlockingGetPort = UVM_PORT_COMMON(UVMBlockingGetPort, UVM_TLM_BLOCKING_GET_MASK,"uvm_blocking_get_port")
UVM_BLOCKING_GET_IMP('m_if', UVMBlockingGetPort)

#class uvm_nonblocking_get_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_NONBLOCKING_GET_MASK,"uvm_nonblocking_get_port")
#  `UVM_NONBLOCKING_GET_IMP(this.m_if, T, t)
#endclass
class UVMNonBlockingGetPort():
    pass
UVMNonBlockingGetPort = UVM_PORT_COMMON(UVMNonBlockingGetPort, UVM_TLM_NONBLOCKING_GET_MASK,"uvm_nonblocking_get_port")
UVM_NONBLOCKING_GET_IMP('m_if', UVMNonBlockingGetPort)

#class uvm_get_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_GET_MASK,"uvm_get_port")
#  `UVM_GET_IMP(this.m_if, T, t)
#endclass
class UVMGetPort():
    pass
UVMGetPort = UVM_PORT_COMMON(UVMGetPort, UVM_TLM_GET_MASK, "uvm_get_port")
UVM_GET_IMP('m_if', UVMGetPort)

#class uvm_blocking_peek_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_BLOCKING_PEEK_MASK,"uvm_blocking_peek_port")
#  `UVM_BLOCKING_PEEK_IMP(this.m_if, T, t)
#endclass
class UVMBlockingPeekPort():
    pass
UVMBlockingPeekPort = UVM_PORT_COMMON(UVMBlockingPeekPort,
        UVM_TLM_BLOCKING_PEEK_MASK, "uvm_blocking_peek_port")
UVM_BLOCKING_PEEK_IMP('m_if', UVMBlockingPeekPort)

#class uvm_nonblocking_peek_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_NONBLOCKING_PEEK_MASK,"uvm_nonblocking_peek_port")
#  `UVM_NONBLOCKING_PEEK_IMP(this.m_if, T, t)
#endclass
class UVMNonBlockingPeekPort():
    pass
UVMNonBlockingPeekPort = UVM_PORT_COMMON(UVMNonBlockingPeekPort,
    UVM_TLM_NONBLOCKING_PEEK_MASK,"uvm_nonblocking_peek_port")
UVM_NONBLOCKING_PEEK_IMP('m_if', UVMNonBlockingPeekPort)

#class uvm_peek_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_PEEK_MASK,"uvm_peek_port")
#  `UVM_PEEK_IMP(this.m_if, T, t)
#endclass
class UVMPeekPort():
    pass
UVMPeekPort = UVM_PORT_COMMON(UVMPeekPort, UVM_TLM_PEEK_MASK,"uvm_peek_port")
UVM_PEEK_IMP('m_if', UVMPeekPort)

#class uvm_blocking_get_peek_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_BLOCKING_GET_PEEK_MASK,"uvm_blocking_get_peek_port")
#  `UVM_BLOCKING_GET_PEEK_IMP(this.m_if, T, t)
#endclass
class UVMBlockingGetPeekPort():
    pass
UVMBlockingGetPeekPort = UVM_PORT_COMMON(UVMBlockingGetPeekPort,
    UVM_TLM_BLOCKING_GET_PEEK_MASK,"uvm_blocking_get_peek_port")
UVM_BLOCKING_GET_PEEK_IMP('m_if', UVMBlockingGetPeekPort)

#class uvm_nonblocking_get_peek_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_NONBLOCKING_GET_PEEK_MASK,"uvm_nonblocking_get_peek_port")
#  `UVM_NONBLOCKING_GET_PEEK_IMP(this.m_if, T, t)
#endclass
class UVMNonBlockingGetPeekPort():
    pass
UVMNonBlockingGetPeekPort = UVM_PORT_COMMON(UVMNonBlockingGetPeekPort,
        UVM_TLM_NONBLOCKING_GET_PEEK_MASK,"uvm_nonblocking_get_peek_port")
UVM_NONBLOCKING_GET_PEEK_IMP('m_if',  UVMNonBlockingGetPeekPort)

#class uvm_get_peek_port #(type T=int)
#  extends uvm_port_base #(uvm_tlm_if_base #(T,T));
#  `UVM_PORT_COMMON(`UVM_TLM_GET_PEEK_MASK,"uvm_get_peek_port")
#  `UVM_GET_PEEK_IMP(this.m_if, T, t)
#endclass
class UVMGetPeekPort():
    pass
UVMGetPeekPort = UVM_PORT_COMMON(UVMGetPeekPort, UVM_TLM_GET_PEEK_MASK,"uvm_get_peek_port")
UVM_GET_PEEK_IMP('m_if',  UVMGetPeekPort)

#------------------------------------------------------------------------------
#
# Class: uvm_*_port #(REQ,RSP)
#
# These bidirectional ports are instantiated by components that ~require~,
# or ~use~, the associated interface to convey transactions. A port can
# be connected to any compatible port, export, or imp port. Unless its
# ~min_size~ is 0, a port ~must~ be connected to at least one implementation
# of its associated interface.
#
# The asterisk in ~uvm_*_port~ is any of the following
#
#|  blocking_transport
#|  nonblocking_transport
#|  transport
#|
#|  blocking_master
#|  nonblocking_master
#|  master
#|
#|  blocking_slave
#|  nonblocking_slave
#|  slave
#
# Ports are connected to interface implementations directly via
# <uvm_*_imp #(REQ,RSP,IMP,REQ_IMP,RSP_IMP)> ports or indirectly via
# hierarchical connections to <uvm_*_port #(REQ,RSP)> and
# <uvm_*_export #(REQ,RSP)> ports.
#
# Type parameters
#
# REQ - The type of request transaction to be communicated by the export
#
# RSP - The type of response transaction to be communicated by the export
#
#------------------------------------------------------------------------------
#
# Function: new
#
# The ~name~ and ~parent~ are the standard <uvm_component> constructor arguments.
# The ~min_size~ and ~max_size~ specify the minimum and maximum number of
# interfaces that must have been supplied to this port by the end of elaboration.
#
#   function new (string name,
#                 uvm_component parent,
#                 int min_size=1,
#                 int max_size=1)
#
#
#class uvm_blocking_master_port #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_PORT_COMMON(`UVM_TLM_BLOCKING_MASTER_MASK,"uvm_blocking_master_port")
#  `UVM_BLOCKING_PUT_IMP(this.m_if, REQ, t)
#  `UVM_BLOCKING_GET_PEEK_IMP(this.m_if, RSP, t)
#endclass
class UVMBlockingMasterPort():  #(type REQ=int, type RSP=REQ)
    pass
UVMBlockingMasterPort = UVM_PORT_COMMON(UVMBlockingMasterPort,
        UVM_TLM_BLOCKING_MASTER_MASK,"uvm_blocking_master_port")
UVM_BLOCKING_PUT_IMP('m_if',  UVMBlockingMasterPort)
UVM_BLOCKING_GET_PEEK_IMP('m_if', UVMBlockingMasterPort)

#class uvm_nonblocking_master_port #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_PORT_COMMON(`UVM_TLM_NONBLOCKING_MASTER_MASK,"uvm_nonblocking_master_port")
#  `UVM_NONBLOCKING_PUT_IMP(this.m_if, REQ, t)
#  `UVM_NONBLOCKING_GET_PEEK_IMP(this.m_if, RSP, t)
#endclass
class UVMNonBlockingMasterPort():   #(type REQ=int, type RSP=REQ)
    pass
UVMNonBlockingMasterPort = UVM_PORT_COMMON(UVMNonBlockingMasterPort,
        UVM_TLM_NONBLOCKING_MASTER_MASK,"uvm_nonblocking_master_port")
UVM_NONBLOCKING_PUT_IMP('m_if',  UVMNonBlockingMasterPort)
UVM_NONBLOCKING_GET_PEEK_IMP('m_if', UVMNonBlockingMasterPort)

#class uvm_master_port #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_PORT_COMMON(`UVM_TLM_MASTER_MASK,"uvm_master_port")
#  `UVM_PUT_IMP(this.m_if, REQ, t)
#  `UVM_GET_PEEK_IMP(this.m_if, RSP, t)
#endclass
class UVMMasterPort():  # (type REQ=int, type RSP=REQ)
    pass
UVMMasterPort = UVM_PORT_COMMON(UVMMasterPort,
        UVM_TLM_MASTER_MASK,"uvm_master_port")
UVM_PUT_IMP('m_if',  UVMMasterPort)
UVM_GET_PEEK_IMP('m_if', UVMMasterPort)

#class uvm_blocking_slave_port #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(RSP, REQ));
#  `UVM_PORT_COMMON(`UVM_TLM_BLOCKING_SLAVE_MASK,"uvm_blocking_slave_port")
#  `UVM_BLOCKING_PUT_IMP(this.m_if, RSP, t)
#  `UVM_BLOCKING_GET_PEEK_IMP(this.m_if, REQ, t)
#endclass
class UVMBlockingSlavePort():  #(type REQ=int, type RSP=REQ)
    pass
UVMBlockingSlavePort = UVM_PORT_COMMON(UVMBlockingSlavePort,
        UVM_TLM_BLOCKING_SLAVE_MASK,"uvm_blocking_slave_port")
UVM_BLOCKING_PUT_IMP('m_if',  UVMBlockingSlavePort)
UVM_BLOCKING_GET_PEEK_IMP('m_if', UVMBlockingSlavePort)

#class uvm_nonblocking_slave_port #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(RSP, REQ));
#  `UVM_PORT_COMMON(`UVM_TLM_NONBLOCKING_SLAVE_MASK,"uvm_nonblocking_slave_port")
#  `UVM_NONBLOCKING_PUT_IMP(this.m_if, RSP, t)
#  `UVM_NONBLOCKING_GET_PEEK_IMP(this.m_if, REQ, t)
#endclass
class UVMNonBlockingSlavePort():  # (type REQ=int, type RSP=REQ)
    pass
UVMNonBlockingSlavePort = UVM_PORT_COMMON(UVMNonBlockingSlavePort,
        UVM_TLM_NONBLOCKING_SLAVE_MASK,"uvm_nonblocking_slave_port")
UVM_NONBLOCKING_PUT_IMP('m_if',  UVMNonBlockingSlavePort)
UVM_NONBLOCKING_GET_PEEK_IMP('m_if', UVMNonBlockingSlavePort)

#class uvm_slave_port #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(RSP, REQ));
#  `UVM_PORT_COMMON(`UVM_TLM_SLAVE_MASK,"uvm_slave_port")
#  `UVM_PUT_IMP(this.m_if, RSP, t)
#  `UVM_GET_PEEK_IMP(this.m_if, REQ, t)
#endclass
class UVMSlavePort():  # (type REQ=int, type RSP=REQ)
    pass
UVMSlavePort = UVM_PORT_COMMON(UVMSlavePort,
        UVM_TLM_SLAVE_MASK,"uvm_slave_port")
UVM_PUT_IMP('m_if',  UVMSlavePort)
UVM_GET_PEEK_IMP('m_if', UVMSlavePort)

#class uvm_blocking_transport_port #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_PORT_COMMON(`UVM_TLM_BLOCKING_TRANSPORT_MASK,"uvm_blocking_transport_port")
#  `UVM_BLOCKING_TRANSPORT_IMP(this.m_if, REQ, RSP, req, rsp)
#endclass
class UVMBlockingTransportPort():   # (type REQ=int, type RSP=REQ)
    pass
UVMBlockingTransportPort = UVM_PORT_COMMON(UVMBlockingTransportPort,
    UVM_TLM_BLOCKING_TRANSPORT_MASK,"uvm_blocking_transport_port")
UVM_BLOCKING_TRANSPORT_IMP('m_if', UVMBlockingTransportPort)

#class uvm_nonblocking_transport_port #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_PORT_COMMON(`UVM_TLM_NONBLOCKING_TRANSPORT_MASK,"uvm_nonblocking_transport_port")
#  `UVM_NONBLOCKING_TRANSPORT_IMP(this.m_if, REQ, RSP, req, rsp)
#endclass
class UVMNonBlockingTransportPort():   # (type REQ=int, type RSP=REQ)
    pass
UVMNonBlockingTransportPort =  UVM_PORT_COMMON(UVMNonBlockingTransportPort,
    UVM_TLM_NONBLOCKING_TRANSPORT_MASK,"uvm_nonblocking_transport_port")
UVM_NONBLOCKING_TRANSPORT_IMP('m_if', UVMNonBlockingTransportPort)


#class uvm_transport_port #(type REQ=int, type RSP=REQ)
#  extends uvm_port_base #(uvm_tlm_if_base #(REQ, RSP));
#  `UVM_PORT_COMMON(`UVM_TLM_TRANSPORT_MASK,"uvm_transport_port")
#  `UVM_TRANSPORT_IMP(this.m_if, REQ, RSP, req, rsp)
#endclass
class UVMTransportPort():  # (type REQ=int, type RSP=REQ)
    pass
UVMTransportPort = UVM_PORT_COMMON(UVMTransportPort,
    UVM_TLM_TRANSPORT_MASK,"uvm_transport_port")
UVM_TRANSPORT_IMP('m_if', UVMTransportPort)



#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.base import *
from uvm.comps import UVMAgent
from uvm.macros import uvm_component_utils

from ubus_master_monitor import ubus_master_monitor
from ubus_master_driver import ubus_master_driver
from ubus_master_sequencer import ubus_master_sequencer

#//------------------------------------------------------------------------------
#//
#// CLASS: ubus_master_agent
#//
#//------------------------------------------------------------------------------


class ubus_master_agent(UVMAgent):
    #
    #  ubus_master_driver driver;
    #  uvm_sequencer#(ubus_transfer) sequencer;
    #  ubus_master_monitor monitor;
    #
    #  // Provide implementations of virtual methods such as get_type_name and create
    #  `uvm_component_utils_begin(ubus_master_agent)
    #    `uvm_field_int(master_id, UVM_DEFAULT)
    #  `uvm_component_utils_end
    #
    #  // new - constructor
    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.master_id = 0
        self.driver = None
        self.sequencer = None
        self.monitor = None

    def build_phase(self, phase):
        super().build_phase(phase)
        self.monitor = ubus_master_monitor.type_id.create("monitor", self)
        if self.get_is_active() == UVM_ACTIVE:
            self.driver = ubus_master_driver.type_id.create("driver", self)
            self.sequencer = ubus_master_sequencer.type_id.create("sequencer",
                    self)

        arr = []
        if UVMConfigDb.get(self, "*", "master_id", arr) is True:
            self.master_id = arr[0]
        UVMConfigDb.set(self, "", "master_id", self.master_id)

    # connect_phase
    def connect_phase(self, phase):
        if self.get_is_active() == UVM_ACTIVE:
            self.driver.seq_item_port.connect(self.sequencer.seq_item_export)
            self.sequencer.addr_ph_port.connect(self.monitor.addr_ph_imp)
        #  endfunction : connect_phase


uvm_component_utils(ubus_master_agent)



#//----------------------------------------------------------------------
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
"""
Contains SystemVerilog (SV) system functions mocked/added
to make the porting SV to Python faster. Also, some functions which are
originally implemented in C-code, are moved here to implement them in
Python. As the backdoor access is implemented via cocotb-object interface,
we don't want to compile vendor-specific C-code for using the backdoor
access. cocotb is already hiding these details from us.
"""

from inspect import getframeinfo, stack
import re
import random
from typing import List, Any

import cocotb

from cocotb_coverage import crv
from cocotb.triggers import Lock, Timer, Combine, First
from cocotb.utils import get_sim_time, simulator
from cocotb_bus.bus import Bus

from .uvm_exceptions import RandomizeError

RET_ERR = 1
RET_OK = 0


def uvm_re_match(rex, _str):
    m = re.search(rex, _str)
    if m is not None:
        return RET_OK
    return RET_ERR


def uvm_glob_to_re(_str: str) -> str:
    """
    Converts a glob-style pattern to Python
    regular expression syntax.

    Returns:
        str: Glob converted to regular expression syntax.
    """
    if _str is None or _str == "":
        return ""
    if _str[0] == "/" and _str[-1] == "/":
        return _str
    # TODO replace * with .*
    res = _str.replace('.', '\\.')
    res = res.replace('*', '.*')
    res = res.replace('[', '\\[')
    res = res.replace(']', '\\]')
    res = res.replace('?', '.')
    # TODO add more substitutions
    return res


def uvm_split_string(_str, sep: str, split_vals: List[str]) -> List[str]:
    res = _str.split(sep)
    for val in res:
        split_vals.append(val)
    return res


class process():

    FINISHED = 0
    RUNNING = 1
    WAITING = 2
    SUSPENDED = 3
    KILLED = 4

    def __init__(self):
        self.status = process.RUNNING
        self.pid = -1


class sv:
    """ Namespace for SV functions """

    default_time_unit = "NS"

    @classmethod
    def clog2(cls, value):
        value = value - 1
        res = 0
        while value > 0:
            value = value >> 1
            res += 1
        return res

    @classmethod
    def is_simulator_active(cls):
        return simulator.is_running()

    @classmethod
    def display(cls, msg, *args):
        form_msg = sv.sformatf(msg, *args)
        print(form_msg)

    @classmethod
    def realtime(cls, unit=None):
        if cls.is_simulator_active() is False:
            return 0
        elif unit is None:
            return get_sim_time(cls.default_time_unit)
        else:
            return get_sim_time(unit)

    @classmethod
    def time(cls, unit=None):
        return cls.realtime(unit)

    @classmethod
    def fwrite(cls, fhandle, msg):
        if hasattr(fhandle, 'write'):
            fhandle.write(msg)
        else:
            print(msg)

    @classmethod
    def isunknown(cls, value):
        # TODO implement this properly
        return False

    @classmethod
    def bits(cls, var):
        if isinstance(var, int):
            return 32
        return 0

    @classmethod
    def fopen(cls, name, mode):
        return open(name, mode)

    @classmethod
    def fclose(cls, handle):
        handle.close()

    @classmethod
    def fdisplay(cls, fhandle, msg, *args):
        form_msg = sv.sformatf(msg, *args)
        form_msg += "\n"  # SV fdisplay behavior
        if fhandle != 0:
            if fhandle != 1:
                fhandle.write(form_msg)
            else:
                print(form_msg)

    @classmethod
    def cast(cls, dest: List, src: Any, T) -> bool:
        if isinstance(src, T):
            if hasattr(dest, 'append'):
                dest.append(src)
            return True
        return False

    STR_RE = re.compile(r'%(\d*[bdshxfpt])')
    formats = ["%b", "%0b", "%0d", "%d", "%s", "%0s", "%h", "%0h", "%f",
            "%p", "%0t", "%t", "%x"]

    @classmethod
    def sscanf(cls, scan_str: str, formats: str, *results):
        """
        Scans the input string with given format pattern.

        Args:
            scan_str (str): Input string to be scanned.

        Returns:
            list: List of matched strings
        """

        res = []
        re_list = get_regex_list(formats)
        acc = ""
        curr_match = ""
        for char in scan_str:
            acc += char
            prefix = re_list[0][2]
            if re_list[0][0].match(acc):
                # Keep adding chars while match holds
                curr_match += char
            elif len(curr_match) > 0:
                # When it stops matching, we're done for curr match
                value = re_list[0][1](curr_match)
                re_list.pop(0)  # Remove matched regex
                res.append(value)  # Append results
                curr_match = ""
                acc = ""
                if len(re_list) == 0:  # We're done, all matched
                    break
            elif (len(acc) >= len(prefix) and acc != prefix):
                acc = ""

        # Special case if last letter is a matching one
        if len(curr_match) > 0:
            # When it stops matching, we're done for curr match
            value = re_list[0][1](curr_match)
            re_list.pop(0)  # Remove matched regex
            res.append(value)  # Append results

        # TODO check from SV spec if all format specifiers must match
        if len(re_list) > 0:  # Regex left unmatched
            return []
        return res

    @classmethod
    def sformatf(cls, msg: str, *args) -> str:
        """
        This is to make porting faster, but should be switched to native python
        formatting inside UVM code

        Args:
            msg (str): String to format containing format specifiers.
            args (*any): Values that are used in formatting.
        Returns:
            str: Formatted string
        """
        #formats = {"%t": "%d", "%0t": "%0d"}
        #for s in formats:
        #    msg = msg.replace(s, formats[s])
        #return sformatf(msg, *args)
        # TODO substitute old types %s/%d etc with {}
        #new_msg = cls.STR_RE.sub(r'{:\1}', msg)
        #print("new_msg is " + new_msg)
        for s in cls.formats:
            if s == "%h" or s == "%0h":
                msg = msg.replace(s, "{:X}")
            else:
                msg = msg.replace(s, "{}")
        return msg.format(*args)

    @classmethod
    def random(cls):
        return random.randint(0, SV_MAX_INT_VALUE)

    @classmethod
    def urandom(cls):
        return random.randint(0, SV_MAX_INT_VALUE)

    @classmethod
    def urandom_range(cls, start, stop):
        return random.randint(start, stop)

    @classmethod
    def sv_assert(cls, val, msg=""):
        if not val:
            caller = getframeinfo(stack()[1][0])
            filename = caller.filename
            line = caller.lineno
            print("$error: {},{} Assertion failed. {}".format(filename, line, msg))

    @classmethod
    def test_plusargs(cls, name):
        arr = []
        sv.value_plusargs(name, arr)
        if len(arr) > 0:
            return True
        return False

    @classmethod
    def value_plusargs(cls, arg_str, arr):
        plusarg_dict = cocotb.plusargs
        for name in plusarg_dict:
            full_arg = name + '='
            for i in range(len(full_arg)):
                if full_arg[i] != arg_str[i]:
                    break
                elif i == len(full_arg) - 1:
                    arr.append(plusarg_dict[name])
                    return plusarg_dict[name]

    @classmethod
    async def fork_join(cls, forks):
        join_list = list(map(lambda t: t.join(), forks))
        await Combine(*join_list)

    @classmethod
    async def fork_join_any(cls, forks):
        join_list = list(map(lambda t: t.join(), forks))
        await First(*join_list)

    @classmethod
    async def fork_join_none(cls, procs):
        res = []
        for proc in procs:
            res.append(cocotb.fork(proc))
        return res


    @classmethod
    def set_bit(cls, int_val: int, idx: int, bit_val=1) -> int:
        if bit_val == 1:
            return int_val | (1 << idx)
        if bit_val == 0:
            return cls.clear_bit(int_val, idx)
        raise ValueError("bit_val must be 0/1. Got: " + str(bit_val))

    @classmethod
    def get_bit(cls, int_val: int, idx: int) -> int:
        return (int_val >> idx) & 0x1

    @classmethod
    def clear_bit(cls, int_val: int, idx: int) -> int:
        return int_val & ~(1 << idx)

    @classmethod
    def get_vector_of_ones(cls, width) -> int:
        return int(''.join(['1' for bit in range(width)]),2)


random.seed(0)

SV_MAX_INT_VALUE = (1 << 31) - 1


class sv_obj(crv.Randomized):
    """ 
    `sv_obj` implement some basic features from SystemVerilog objects like
    constrained randomisation and random stability. Each object maintains
    its own random state, and thus calls to `randomize` result in the 
    same sequence of values, given the same seed.
    """

    num_objs = 0

    def __init__(self):
        crv.Randomized.__init__(self)
        self._sv_seed = sv.urandom()
        self._sv_rand_obj = []
        random.seed(self._sv_seed)
        self._sv_rand_state = random.getstate()
        self._rand_mode = True
        sv_obj.num_objs += 1


    def constraint(self, c):
        """
        Adds a constraint into the object.

        Args:
            c (constraint): Lambda expression or function.

        """
        self.add_constraint(c)


    def rand_mode(self, mode):
        if mode:
            self._rand_mode = True
        else:
            self._rand_mode = False


    def rand(self, key, val_list=None):
        """
        Mark given class property as randomized. A call to `sv_obj.randomize()`
        or `sv_obj.randomize_with()` will then randomize this value.

        Args:
            key (str): Name of the property.
            val_list (list|range): Optional constraints for randomisation.
        """
        if hasattr(key, "randomize"):
            if val_list is None:
                self._sv_rand_obj.append(key)
        elif hasattr(self, key) and val_list is None:
            if self.is_valid_rand_type(key):
                self._sv_rand_obj.append(key)
            else:
                raise TypeError('Attr {} of {} requires range or value list for rand()'
                    .format(key, type(getattr(self, key))))
        else:
            self.add_rand(key, val_list)

    def is_valid_rand_type(self, key):
        return not isinstance(getattr(self, key), (int, str, float))

    def pre_randomize(self):
        """
        Classes derived from `sv_obj` can implement this callback, which is
        called before the actual randomization in `randomize` or
        `randomize_with`.
        """
        pass

    def post_randomize(self):
        """
        Classes derived from `sv_obj` can implement this callback, which is
        called after the actual randomization in `randomize` or
        `randomize_with`.
        """
        pass

    def randomize(self, recurse=True):
        """
        Randomizes properties in object marked with rand(). Recurses to sub-objects
        randomizing them as well.

        raises:
            RandomizeError
        """
        if self._rand_mode is False:
            return True
        try:
            random.setstate(self._sv_rand_state)
            self.pre_randomize()
            ok = True
            if recurse:
                for entry in self._sv_rand_obj:
                    obj = entry
                    if hasattr(self, entry):
                        obj = getattr(self, entry)
                    ok = ok and obj.randomize()
            super().randomize()
            self.post_randomize()
            self._sv_rand_state = random.getstate()
            return True and ok
        except Exception as e:
            raise RandomizeError('randomize() failed for ' + str(self) + '\n' +
                    str(e))


    def randomize_with(self, *constr):
        """
        Randomizes the object with inline constraints.

        An example call::

            pkt = Packet('packet')
            pkt.randomize_with(
                lambda addr: addr in [0, 1, 2, 3],
                lambda data: data > 0
            )

        raises:
            RandomizeError

        """
        if self._rand_mode is False:
            return True
        try:
            self.pre_randomize()
            ok = True
            for entry in self._sv_rand_obj:
                obj = entry
                if hasattr(self, entry):
                    obj = getattr(self, entry)
                ok = ok and obj.randomize()
            super().randomize_with(*constr)
            self.post_randomize()
            return True and ok
        except Exception as e:
            raise RandomizeError(
                'randomize_with() failed for ' + str(self) + '\n' + str(e))


class semaphore():

    def __init__(self, count=1):
        self.max_count = count
        self.count = count
        self.lock = Lock("sem_lock")
        self.locked = False


    async def get(self, count=1):
        if self.count > count:
            self.count -= count
            await Timer(0, "NS")
        elif count > self.max_count:
            raise Exception("Tried to get {} > max count {}".format(count,
                self.max_count))
        else:
            await self.lock.acquire()
            self.locked = True
            self.count -= count

    def put(self, count=1):
        if self.count < self.max_count:
            self.count += count
            if self.locked is True:
                self.locked = False
                self.lock.release()

    def try_get(self, count=1):
        if self.count >= count:
            self.count -= count
            return True
        return False


class sv_if(Bus):

    def __init__(self, entity, name, signals,
            optional_signals=[], bus_separator="_", array_idx=None):
        Bus.__init__(self, entity, name, signals, optional_signals,
                bus_separator, array_idx)


def sformatf(fmt, *args):
    return fmt % args


def cat(*args):
    ret = ""
    for a in args:
        ret += a

    return ret



async def wait(cond, ev):
    if not callable(cond):
        raise Exception("wait expects the first arguments to be callable")

    while True:
        if cond():
            break
        else:
            await ev.wait()
            ev.clear()

# Each letter matches a pair:
#   1. Regex to parse that value from string
#   2. Lambda function to convert the value
RE_FORMATS = {
    "d": [re.compile(r'^(-?[0-9_]+)$'), lambda s: int(s, 10), '-'],
    "h": [re.compile(r'^(0x[0-9a-fA-F_]+)$'), lambda s: int(s, 16), '0x'],
    "b": [re.compile(r'^([01_]+)$'), lambda s: int(s, 2), ''],
    "o": [re.compile(r'^(o[01_]+)$'), lambda s: int(s, 8), 'o'],
    "f": [re.compile(r'^(-?[0-9]*\.?[0-9]*)$'), float, '-'],
    "s": [re.compile(r'^(\w+)$'), lambda s: s, '']
}


def get_regex_list(formats):
    acc = ""
    re_format = re.compile("^%([dshbfo])$")
    results = []
    for char in formats:
        acc += char
        if len(acc) == 2:
            match = re_format.match(acc)
            if match:
                if match[1] in RE_FORMATS:
                    results.append(RE_FORMATS[match[1]])
                    acc = ""
                else:
                    raise Exception("Format spec {} not supported".format(match))
            else:
                acc = acc[1]  # Discard 1st character out of 2
    return results




import unittest
from uvm.base.uvm_config_db import UVMConfigDb, UVMConfigDbOptions
from uvm.base.uvm_debug import UVMDebug

str1 = "uvm_test_top.ubus_example_tb0.ubus0.masters[0].monitor"

#UVMDebug.DEBUG = True


class TestUVMConfigDb(unittest.TestCase):

    def test_exists(self):
        pass
        # ok = True
        # self.assertEqual(ok, False)
        # ok = False
        # self.assertEqual(ok, True)

    def test_set_get(self):
        cntxt = None
        inst_name = "my_comp"
        field_name = "field1"
        value = 123
        UVMConfigDb.set(cntxt, "other_name", field_name, 666, T=None)
        UVMConfigDb.set(cntxt, inst_name, field_name, value, T=None)
        UVMConfigDb.set(cntxt, inst_name + ".*", field_name, 555, T=None)
        got_val = []
        ok = UVMConfigDb.get(cntxt, inst_name, field_name, got_val, T=None)
        self.assertEqual(ok, True)
        self.assertEqual(got_val[0], value)
        got_val = []
        ok = UVMConfigDb.get(cntxt, inst_name, "wrong_name", got_val, T=None)
        self.assertEqual(ok, False)
        self.assertEqual(len(got_val), 0)

        hier_name = inst_name + "." + "child"
        got_val = []
        ok = UVMConfigDb.get(cntxt, hier_name, field_name, got_val, T=None)
        self.assertEqual(ok, True)
        self.assertEqual(got_val[0], 555)


    def test_hier_set_get(self):
        UVMConfigDbOptions.tracing = True
        UVMConfigDb.set(None, str1, "master_id", 666, T=None)

        arr = []
        ok = UVMConfigDb.get(None, str1, "master_id", arr)
        self.assertEqual(ok, True)
        self.assertEqual(arr[0], 666)


    def test_set_override(self):
        pass
        # self.assertEqual(0, 1)


if __name__ == '__main__':
    unittest.main()




import unittest
from uvm.reg import (UVMRegMap, UVMRegBlock, UVMReg, UVMMem)
from uvm.reg.uvm_reg_model import *


class TestUVMRegMap(unittest.TestCase):


    def test_map_base(self):
        reg_map = UVMRegMap()
        reg_blk = UVMRegBlock('my_blk')
        reg_map.configure(reg_blk, base_addr=0, n_bytes=32, endian=0,
                byte_addressing=1)
        reg_blk.add_map(reg_map)
        r1 = UVMReg('r1', 32, False)
        r1.configure(reg_blk)
        reg_map.add_reg(r1, 4)
        reg_blk.lock_model()
        self.assertEqual(r1.get_offset(), 4)
        r1.set_offset(reg_map, 16)
        self.assertEqual(r1.get_offset(), 16)
        reg_map.reset(kind="SOFT")


    def test_add_parent_map(self):
        pass


    def test_get_physical_addresses(self):
        # reg_map = UVMRegMap()
        reg_blk = UVMRegBlock('my_blk2')
        reg_map = reg_blk.create_map("default_map2", 0x0, 4,
                UVM_LITTLE_ENDIAN)
        reg_blk.default_map = reg_map

        #reg_blk.add_map(reg_map)
        mem = UVMMem("my_mem_inst", 1024, 32, "RW")
        mem.configure(reg_blk, "")

        reg_map.m_system_n_bytes = 4
        reg_map.m_n_bytes = 4
        reg_map.add_mem(mem, 0x2000, "RW")
        reg_map.Xinit_address_mapX()

        mem_offset = 8192
        nbytes = 4
        addrs = []
        bus_width = reg_map.get_physical_addresses(mem_offset, 0, nbytes, addrs)
        self.assertEqual(addrs[0], mem_offset)

        mem_offset += 4
        bus_width = reg_map.get_physical_addresses(mem_offset, 0, nbytes, addrs)
        self.assertEqual(addrs[0], mem_offset)
        self.assertEqual(bus_width, 4)


    def test_error_conditions(self):
        reg_blk = UVMRegBlock('my_blk2')
        reg_map = reg_blk.create_map("default_map2", 0x0, 4,
            UVM_LITTLE_ENDIAN)
        reg_map.set_sequencer(None, None)
        reg_map.set_submap_offset(None, 0x123)
        err = reg_map.get_submap_offset(None)
        self.assertEqual(err, -1)



#//
#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc. 
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2010 Paradigm-works, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#
#`ifndef UVM_NO_DEPRECATED
#
#//------------------------------------------------------------------------------
#// Title- Predefined Sequences
#//
#// This section defines several sequences that ship with the UVM library.
#//------------------------------------------------------------------------------
#
#//------------------------------------------------------------------------------
#//
#// CLASS- uvm_random_sequence
#//
#// This sequence randomly selects and executes a sequence from the sequencer's
#// sequence library, excluding uvm_random_sequence itself, and
#// uvm_exhaustive_sequence. 
#//
#// The uvm_random_sequence class is a built-in sequence that is preloaded into
#// every sequencer's sequence library with the name "uvm_random_sequence".
#//
#// The number of selections and executions is determined by the count property
#// of the sequencer (or virtual sequencer) on which uvm_random_sequence is
#// operating. See <uvm_sequencer_base> for more information.
#//
#//------------------------------------------------------------------------------
#
#class uvm_random_sequence extends uvm_sequence #(uvm_sequence_item);
#
#  rand protected int unsigned l_count=-1;
#  local int unsigned l_exhaustive_seq_kind;
#  local int unsigned max_kind;
#  rand local int unsigned l_kind;
#  protected bit m_success;
#
#// Function- get_count
#//
#// Returns the count of the number of sub-sequences which are randomly generated.
#// By default, count is equal to the value from the sequencer's count variable.
#// However, if the sequencer's count variable is -1, then a random value between
#// 0 and sequencer.max_random_count (exclusive) is chosen. The sequencer's
#// count variable is subsequently reset to the random value that was used. If
#// get_count() is call before the sequence has started, the return value will
#// be sequencer.count, which may be -1.
#
#function int unsigned get_count();
#  if(l_count == -1)
#    return m_sequencer.count;
#  return l_count;
#endfunction
#
#// new
#// ---
#
#function new(string name="uvm_random_sequence");
#  super.new(name);
#endfunction
#
#
#// body
#// ----
#
#task body();
#  pick_sequence.constraint_mode(0);
#  if (m_sequencer.count == -1) begin
#    if (!randomize(l_count) with { l_count > 0 &&
#                                   l_count < m_sequencer.max_random_count; })
#      uvm_report_fatal("RANDSEQ", "Randomization for l_count failed in random sequence body", UVM_NONE);
#    m_sequencer.count = l_count;
#  end
#  else
#    l_count = m_sequencer.count;
#  max_kind = m_sequencer.sequences.size();
#  l_exhaustive_seq_kind = m_sequencer.get_seq_kind("uvm_exhaustive_sequence");
#  repeat (l_count) begin
#    if (!randomize(l_kind) with { l_kind > l_exhaustive_seq_kind && 
#      l_kind < max_kind; })
#      uvm_report_fatal("RANDSEQ", "Randomization for l_kind failed in random sequence body", UVM_NONE);
#    do_sequence_kind(l_kind);
#  end
#  m_sequencer.m_random_count++;
#  pick_sequence.constraint_mode(1);
#endtask 
#
#
#//Implement data functions
#function void do_copy (uvm_object rhs);
#  uvm_random_sequence seq;
#  if(rhs==null) return;
#  if(!$cast(seq, rhs)) return;
#  l_count = seq.l_count;
#endfunction
#
#function bit do_compare (uvm_object rhs, uvm_comparer comparer);
#  uvm_random_sequence seq;
#  do_compare = 1;
#  if(rhs==null) return 0;
#  if(!$cast(seq, rhs)) return 0;
#  do_compare &= comparer.compare_field_int("l_count", l_count, seq.l_count, 
#    $bits(l_count));
#endfunction
#
#function void do_print (uvm_printer printer);
#  printer.print_field_int("l_count", l_count, $bits(l_count));
#endfunction
#
#function void do_record (uvm_recorder recorder);
#  recorder.record_field_int("l_count", l_count, $bits(l_count));
#endfunction // void
#
#  function uvm_object create (string name="");
#    uvm_random_sequence i; i=new(name);
#    return i;
#  endfunction
#
#  virtual function string get_type_name();
#     return "uvm_random_sequence";
#  endfunction
#
#  // Macro for factory creation
#  `uvm_object_registry(uvm_random_sequence, "uvm_random_sequence")
#
#endclass
#
#
#//------------------------------------------------------------------------------
#//
#// CLASS- uvm_exhaustive_sequence
#//
#// This sequence randomly selects and executes each sequence from the
#// sequencer's sequence library once, excluding itself and uvm_random_sequence.
#//
#// The uvm_exhaustive_sequence class is a built-in sequence that is preloaded 
#// into every sequencer's sequence library with the name
#// "uvm_exaustive_sequence".
#//
#//------------------------------------------------------------------------------
#
#class uvm_exhaustive_sequence extends uvm_sequence #(uvm_sequence_item);
#
#  rand protected int unsigned l_count;
#  local int unsigned l_exhaustive_seq_kind;
#  local int unsigned max_kind;
#  randc local bit[9:0] l_kind;
#  protected bit m_success;
#
#// new
#// ---
#
#function new(string name="uvm_exhaustive_sequence");
#  super.new(name);
#endfunction
#
#
#// body
#// ----
#
#task body();
#    int i;
#    
#  pick_sequence.constraint_mode(0);
#    //$display("In exhaustive sequence body, num_sequences: %0d", m_sequencer.num_sequences());
#
#
#    //for (i = 0; i < m_sequencer.num_sequences(); i++ ) begin
#    //  $display("seq: %0d: %s", i, m_sequencer.sequences[i]);
#    //  $display("get_seq_kind[%s]: %0d", m_sequencer.sequences[i], get_seq_kind(m_sequencer.sequences[i]));
#    //end
#    
#    
#  l_count = m_sequencer.sequences.size() - 2;
#  max_kind = m_sequencer.sequences.size();
#  l_exhaustive_seq_kind = m_sequencer.get_seq_kind("uvm_exhaustive_sequence");
#  repeat (l_count) begin
#    if (!randomize(l_kind) with { l_kind > l_exhaustive_seq_kind; 
#      l_kind < max_kind; }) // l_kind is randc
#      uvm_report_fatal("RANDSEQ", "Randomization for l_kind failed in exhaustive sequence body", UVM_NONE);
#
#    //$display ("Chosen l_kind: %0d", l_kind);
#    do_sequence_kind(l_kind);
#  end
#  m_sequencer.m_exhaustive_count++;
#  pick_sequence.constraint_mode(1);
#endtask 
#
#
#//Implement data functions
#function void do_copy (uvm_object rhs);
#  uvm_exhaustive_sequence seq;
#  if(rhs==null) return;
#  if(!$cast(seq, rhs)) return;
#  l_count = seq.l_count;
#endfunction
#
#function bit do_compare (uvm_object rhs, uvm_comparer comparer);
#  uvm_exhaustive_sequence seq;
#  do_compare = 1;
#  if(rhs==null) return 0;
#  if(!$cast(seq, rhs)) return 0;
#  do_compare &= comparer.compare_field_int("l_count", l_count, seq.l_count, 
#    $bits(l_count));
#endfunction
#
#function void do_print (uvm_printer printer);
#  printer.print_field_int("l_count", l_count, $bits(l_count));
#endfunction
#
#function void do_record (uvm_recorder recorder);
#  recorder.record_field_int("l_count", l_count, $bits(l_count));
#endfunction // void
#
#function uvm_object create (string name="");
#  uvm_exhaustive_sequence i; i=new(name);
#  return i;
#endfunction
#
#virtual function string get_type_name();
#   return "uvm_exhaustive_sequence";
#endfunction
#
#// Macro for factory creation
#`uvm_object_registry(uvm_exhaustive_sequence, "uvm_exhaustive_sequence")
#
#endclass
#
#
#//------------------------------------------------------------------------------
#//
#// CLASS- uvm_simple_sequence
#//
#// This sequence simply executes a single sequence item. 
#//
#// The item parameterization of the sequencer on which the uvm_simple_sequence
#// is executed defines the actual type of the item executed.
#//
#// The uvm_simple_sequence class is a built-in sequence that is preloaded into
#// every sequencer's sequence library with the name "uvm_simple_sequence". 
#//
#// See <uvm_sequencer #(REQ,RSP)> for more information on running sequences.
#//
#//------------------------------------------------------------------------------
#
#class uvm_simple_sequence extends uvm_sequence #(uvm_sequence_item);
#
#  protected rand uvm_sequence_item item;
#
#  // new
#  // ---
#  function new (string name="uvm_simple_sequence");
#    super.new(name);
#    // Initialized to avoid potential warnings if this class instance 
#    // is randomized without calling its body()
#    item = new;
#  endfunction
#
#  // body
#  // ----
#  task body();
#    `uvm_do(item)
#    m_sequencer.m_simple_count++;
#  endtask
#
#  function uvm_object create (string name="");
#    uvm_simple_sequence i;
#    i=new(name);
#    return i;
#  endfunction
#
#  virtual function string get_type_name();
#     return "uvm_simple_sequence";
#  endfunction
#
#  // Macro for factory creation
#  `uvm_object_registry(uvm_simple_sequence, "uvm_simple_sequence")
#
#endclass
#
#`endif // UVM_NO_DEPRECATED



#
#-----------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#-----------------------------------------------------------------------------

from .sv import sv
from .uvm_recorder import UVMRecorder
from .uvm_object import UVMObject
from .uvm_pool import UVMEventPool


class UVMTransaction(UVMObject):
    """
    The `UVMTransaction` class is the root base class for UVM transactions.
    Inheriting all the methods of `UVMObject`, `UVMTransaction` adds a timing and
    recording interface.

    This class provides timestamp properties, notification events, and transaction
    recording support.

    Use of this class as a base for user-defined transactions
    is deprecated. Its subtype, `uvm_sequence_item`, shall be used as the
    base class for all user-defined transaction types.

    The intended use of this API is via a `uvm_driver` to call `accept_tr`,
    `begin_tr`, and `end_tr` during the course of
    sequence item execution. These methods in the component base class will
    call into the corresponding methods in this class to set the corresponding
    timestamps (`accept_time`, `begin_time`, and `end_time`), trigger the
    corresponding event (`begin_event` and `end_event`, and, if enabled,
    record the transaction contents to a vendor-specific transaction database.

    Note that get_next_item/item_done when called on a `uvm_seq_item_pull_port`
    will automatically trigger the `begin_event` and `end_event` via calls to `begin_tr` and `end_tr`.
    While convenient, it is generally the responsibility of drivers to mark a
    transaction's progress during execution.  To allow the driver or layering sequence
    to control sequence item timestamps, events, and recording, you must call
    `UVM_SEQ_ITEM_PULL_IMP.disable_auto_item_recording` at the beginning
    of the driver's `run_phase` task.

    Users may also use the transaction's event pool, `events`,
    to define custom events for the driver to trigger and the sequences to wait on. Any
    in-between events such as marking the beginning of the address and data
    phases of transaction execution could be implemented via the
    `events` pool.

    In pipelined protocols, the driver may release a sequence (return from
    `finish_item` or it's `uvm_do` macro) before the item has been completed.
    If the driver uses the `begin_tr`/`end_tr` API in `UVMComponent`, the sequence can
    wait on the item's `end_event` to block until the item was fully executed,
    as in the following example.

    .. code-block:: systemverilog

        task uvm_execute(item, ...)
            // can use the `uvm_do macros as well
            start_item(item)
            item.randomize()
            finish_item(item)
            item.self.end_event.wait_on()
            // get_response(rsp, item.get_transaction_id()); //if needed
        endtask


    A simple two-stage pipeline driver that can execute address and
    data phases concurrently might be implemented as follows:

    .. code-block:: systemverilog

        task run()

            // this driver supports a two-deep pipeline
            fork
              do_item()
              do_item()
            join
        endtask


        task do_item()

          forever begin
            mbus_item req

            lock.get()

            seq_item_port.get(req); // Completes the sequencer-driver handshake

            accept_tr(req)

              // request bus, wait for grant, etc.

            begin_tr(req)

              // execute address phase

            // allows next transaction to begin address phase
            lock.put()

              // execute data phase
              // (may trigger custom "data_phase" event here)

            end_tr(req)

          end

        endtask: do_item
    """

    #  // Variable: events
    #  //
    #  // The event pool instance for this transaction. This pool is used to track
    #  // various milestones: by default, begin, accept, and end
    #
    #  const uvm_event_pool events = new
    #
    #
    #  // Variable: self.begin_event
    #  //
    #  // A ~uvm_event#(uvm_object)~ that is triggered when this transaction's actual execution on the
    #  // bus begins, typically as a result of a driver calling <uvm_component::begin_tr>.
    #  // Processes that wait on this event will block until the transaction has
    #  // begun.
    #  //
    #  // For more information, see the general discussion for <uvm_transaction>.
    #  // See <uvm_event#(T)> for details on the event API.
    #  //
    #  uvm_event#(uvm_object) self.begin_event
    #
    #  // Variable: self.end_event
    #  //
    #  // A ~uvm_event#(uvm_object)~ that is triggered when this transaction's actual execution on
    #  // the bus ends, typically as a result of a driver calling <uvm_component::end_tr>.
    #  // Processes that wait on this event will block until the transaction has
    #  // ended.
    #  //
    #  // For more information, see the general discussion for <uvm_transaction>.
    #  // See <uvm_event#(T)> for details on the event API.
    #  //
    #  //| virtual task my_sequence::body()
    #  //|  ...
    #  //|  start_item(item);    \
    #  //|  item.randomize();     } `uvm_do(item)
    #  //|  finish_item(item);   /
    #  //|  // return from finish item does not always mean item is completed
    #  //|  item.self.end_event.wait_on()
    #  //|  ...
    #  //
    #  uvm_event#(uvm_object) self.end_event

    # Function: new
    #
    # Creates a new transaction object. The name is the instance name of the
    # transaction. If not supplied, then the object is unnamed.
    #
    def __init__(self, name="", initiator=None):
        UVMObject.__init__(self, name)
        self.initiator = initiator
        self.m_transaction_id = -1
        self.events = UVMEventPool()
        self.begin_event = self.events.get("begin")
        self.end_event = self.events.get("end")
        self.stream_handle = None  # For recording
        self.tr_recorder = None

        self.begin_time = -1
        self.end_time = -1
        self.accept_time = -1


    #  // Function: accept_tr
    #  //
    #  // Calling ~accept_tr~ indicates that the transaction item has been received by
    #  // a consumer component. Typically a <uvm_driver #(REQ,RSP)> would call <uvm_component::accept_tr>,
    #  // which calls this method-- upon return from a ~get_next_item()~, ~get()~, or ~peek()~
    #  // call on its sequencer port, <uvm_driver#(REQ,RSP)::seq_item_port>.
    #  //
    #  // With some
    #  // protocols, the received item may not be started immediately after it is
    #  // accepted. For example, a bus driver, having accepted a request transaction,
    #  // may still have to wait for a bus grant before beginning to execute
    #  // the request.
    #  //
    #  // This function performs the following actions:
    #  //
    #  // - The transaction's internal accept time is set to the current simulation
    #  //   time, or to accept_time if provided and non-zero. The ~accept_time~ may be
    #  //   any time, past or future.
    #  //
    #  // - The transaction's internal accept event is triggered. Any processes
    #  //   waiting on the this event will resume in the next delta cycle.
    #  //
    #  // - The <do_accept_tr> method is called to allow for any post-accept
    #  //   action in derived classes.
    #
    #  extern function void accept_tr (time accept_time = 0)
    def accept_tr(self, accept_time=0):
        e = None
        if accept_time != 0:
            self.accept_time = accept_time
        else:
            self.accept_time = sv.realtime()

        self.do_accept_tr()
        e = self.events.get("accept")
        if e is not None:
            e.trigger()


    #  // Function: do_accept_tr
    #  //
    #  // This user-definable callback is called by <accept_tr> just before the accept
    #  // event is triggered. Implementations should call ~super.do_accept_tr~ to
    #  // ensure correct operation.
    #
    #  extern virtual protected function void do_accept_tr ()
    def do_accept_tr(self):
        return

    #
    #  // Function: begin_tr
    #  //
    #  // This function indicates that the transaction has been started and is not
    #  // the child of another transaction. Generally, a consumer component begins
    #  // execution of a transactions it receives.
    #  //
    #  // Typically a <uvm_driver #(REQ,RSP)> would call <uvm_component::begin_tr>, which
    #  // calls this method, before actual execution of a sequence item transaction.
    #  // Sequence items received by a driver are always a child of a parent sequence.
    #  // In this case, begin_tr obtains the parent handle and delegates to <begin_child_tr>.
    #  //
    #  // See <accept_tr> for more information on how the
    #  // begin-time might differ from when the transaction item was received.
    #  //
    #  // This function performs the following actions:
    #  //
    #  // - The transaction's internal start time is set to the current simulation
    #  //   time, or to begin_time if provided and non-zero. The begin_time may be
    #  //   any time, past or future, but should not be less than the accept time.
    #  //
    #  // - If recording is enabled, then a new database-transaction is started with
    #  //   the same begin time as above.
    #  //
    #  // - The <do_begin_tr> method is called to allow for any post-begin action in
    #  //   derived classes.
    #  //
    #  // - The transaction's internal begin event is triggered. Any processes
    #  //   waiting on this event will resume in the next delta cycle.
    #  //
    #  // The return value is a transaction handle, which is valid (non-zero) only if
    #  // recording is enabled. The meaning of the handle is implementation specific.
    #
    #
    #  extern function integer begin_tr (time begin_time = 0)
    def begin_tr(self, begin_time=0):
        return self.m_begin_tr(begin_time)

    #
    #  // Function: begin_child_tr
    #  //
    #  // This function indicates that the transaction has been started as a child of
    #  // a parent transaction given by ~parent_handle~. Generally, a consumer
    #  // component calls this method via <uvm_component::begin_child_tr> to indicate
    #  // the actual start of execution of this transaction.
    #  //
    #  // The parent handle is obtained by a previous call to begin_tr or
    #  // begin_child_tr. If the parent_handle is invalid (=0), then this function
    #  // behaves the same as <begin_tr>.
    #  //
    #  // This function performs the following actions:
    #  //
    #  // - The transaction's internal start time is set to the current simulation
    #  //   time, or to begin_time if provided and non-zero. The begin_time may be
    #  //   any time, past or future, but should not be less than the accept time.
    #  //
    #  // - If recording is enabled, then a new database-transaction is started with
    #  //   the same begin time as above. The inherited <uvm_object::record> method
    #  //   is then called, which records the current property values to this new
    #  //   transaction. Finally, the newly started transaction is linked to the
    #  //   parent transaction given by parent_handle.
    #  //
    #  // - The <do_begin_tr> method is called to allow for any post-begin
    #  //   action in derived classes.
    #  //
    #  // - The transaction's internal begin event is triggered. Any processes
    #  //   waiting on this event will resume in the next delta cycle.
    #  //
    #  // The return value is a transaction handle, which is valid (non-zero) only if
    #  // recording is enabled. The meaning of the handle is implementation specific.
    #
    #  extern function integer begin_child_tr (time begin_time = 0,
    #                                               integer parent_handle = 0)
    #Use a parent handle of zero to link to the parent after begin
    def begin_child_tr(self, begin_time=0, parent_handle=0):
        return self.m_begin_tr(begin_time, parent_handle)

    #
    #
    #  // Function: do_begin_tr
    #  //
    #  // This user-definable callback is called by <begin_tr> and <begin_child_tr> just
    #  // before the begin event is triggered. Implementations should call
    #  // ~super.do_begin_tr~ to ensure correct operation.
    #
    #  extern virtual protected function void do_begin_tr ()
    def do_begin_tr(self):
        pass


    #  // Function: end_tr
    #  //
    #  // This function indicates that the transaction execution has ended.
    #  // Generally, a consumer component ends execution of the transactions it
    #  // receives.
    #  //
    #  // You must have previously called <begin_tr> or <begin_child_tr> for this
    #  // call to be successful.
    #  //
    #  // Typically a <uvm_driver #(REQ,RSP)> would call <uvm_component::end_tr>, which
    #  // calls this method, upon completion of a sequence item transaction.
    #  // Sequence items received by a driver are always a child of a parent sequence.
    #  // In this case, begin_tr obtain the parent handle and delegate to <begin_child_tr>.
    #  //
    #  // This function performs the following actions:
    #  //
    #  // - The transaction's internal end time is set to the current simulation
    #  //   time, or to ~end_time~ if provided and non-zero. The ~end_time~ may be any
    #  //   time, past or future, but should not be less than the begin time.
    #  //
    #  // - If recording is enabled and a database-transaction is currently active,
    #  //   then the record method inherited from uvm_object is called, which records
    #  //   the final property values. The transaction is then ended. If ~free_handle~
    #  //   is set, the transaction is released and can no longer be linked to (if
    #  //   supported by the implementation).
    #  //
    #  // - The <do_end_tr> method is called to allow for any post-end
    #  //   action in derived classes.
    #  //
    #  // - The transaction's internal end event is triggered. Any processes waiting
    #  //   on this event will resume in the next delta cycle.
    #
    #  extern function void end_tr (time end_time=0, bit free_handle=1)
    def end_tr(self, end_time=0, free_handle=1):
        self.end_time = end_time
        if end_time == 0:
            self.end_time = sv.realtime()
        self.do_end_tr()  # Callback prior to actual ending of transaction

        if(self.is_recording_enabled() and (self.tr_recorder is not None)):
            self.record(self.tr_recorder)
            self.tr_recorder.close(self.end_time)

            if free_handle:
                # once freed, can no longer link to
                self.tr_recorder.free()
        self.tr_recorder = None

        self.end_event.trigger()


    #  // Function: do_end_tr
    #  //
    #  // This user-definable callback is called by <end_tr> just before the end event
    #  // is triggered. Implementations should call ~super.do_end_tr~ to ensure correct
    #  // operation.
    #
    #  extern virtual protected function void do_end_tr ()
    def do_end_tr(self):
        return


    #  // Function: get_tr_handle
    #  //
    #  // Returns the handle associated with the transaction, as set by a previous
    #  // call to <begin_child_tr> or `begin_tr` with transaction recording enabled.
    #
    #  extern function integer get_tr_handle ()
    def get_tr_handle (self):
        if self.tr_recorder is not None:
            return self.tr_recorder.get_handle()
        else:
            return 0


    #  // Function: disable_recording
    #  //
    #  // Turns off recording for the transaction stream. This method does not
    #  // effect a <uvm_component>'s recording streams.
    def disable_recording(self):
        self.stream_handle = None


    #  // Function: enable_recording
    #  // Turns on recording to the ~stream~ specified.
    #  //
    #  // If transaction recording is on, then a call to ~record~ is made when the
    #  // transaction is ended.
    def enable_recording(self, stream):
        self.stream_handle = stream


    #  // Function: is_recording_enabled
    #  //
    #  // Returns 1 if recording is currently on, 0 otherwise.
    #
    #  extern function bit is_recording_enabled()
    def is_recording_enabled(self):
        return self.stream_handle is not None

    #  // Function: is_active
    #  //
    #  // Returns 1 if the transaction has been started but has not yet been ended.
    #  // Returns 0 if the transaction has not been started.
    #
    #  extern function bit is_active ()
    def is_active(self):
        return self.end_time == -1


    #  // Function: get_event_pool
    #  //
    #  // Returns the event pool associated with this transaction.
    #  //
    #  // By default, the event pool contains the events: begin, accept, and end.
    #  // Events can also be added by derivative objects. An event pool is a
    #  // specialization of <uvm_pool#(KEY,T)>, e.g. a ~uvm_pool#(uvm_event)~.
    #
    #  extern function uvm_event_pool get_event_pool ()
    def get_event_pool(self):
        return self.events


    #  // Function: set_initiator
    #  //
    #  // Sets initiator as the initiator of this transaction.
    #  //
    #  // The initiator can be the component that produces the transaction. It can
    #  // also be the component that started the transaction. This or any other
    #  // usage is up to the transaction designer.
    #
    #  extern function void set_initiator (uvm_component initiator)
    def set_initiator(self, initiator):
        self.initiator = initiator


    #  // Function: get_initiator
    #  //
    #  // Returns the component that produced or started the transaction, as set by
    #  // a previous call to set_initiator.
    #  extern function uvm_component get_initiator ()


    #  // Function: get_accept_time
    #
    def get_accept_time(self):
        return self.accept_time

    #  // Function: get_begin_time
    #
    def get_begin_time(self):
        return self.begin_time


    #  // Function: get_end_time
    #  //
    #  // Returns the time at which this transaction was accepted, begun, or ended,
    #  // as by a previous call to <accept_tr>, <begin_tr>, <begin_child_tr>, or <end_tr>.
    #
    #  extern function time   get_end_time       ()
    def get_end_time(self):
        return self.end_time


    #  // Function: set_transaction_id
    #  //
    #  // Sets this transaction's numeric identifier to id. If not set via this
    #  // method, the transaction ID defaults to -1.
    #  //
    #  // When using sequences to generate stimulus, the transaction ID is used along
    #  // with the sequence ID to route responses in sequencers and to correlate
    #  // responses to requests.
    def set_transaction_id(self, id):
        self.m_transaction_id = id

    #  // Function: get_transaction_id
    #  //
    #  // Returns this transaction's numeric identifier, which is -1 if not set
    #  // explicitly by ~set_transaction_id~.
    #  //
    #  // When using a <uvm_sequence #(REQ,RSP)> to generate stimulus, the transaction
    #  // ID is used along
    #  // with the sequence ID to route responses in sequencers and to correlate
    #  // responses to requests.
    def get_transaction_id(self):
        return self.m_transaction_id

    #  //----------------------------------------------------------------------------
    #  //
    #  // Internal methods properties; do not use directly
    #  //
    #  //----------------------------------------------------------------------------

    #  //Override data control methods for internal properties
    #  extern virtual function void do_print  (uvm_printer printer)

    #  extern virtual function void do_record (uvm_recorder recorder)

    #  extern virtual function void do_copy   (uvm_object rhs)

    #  extern protected function integer m_begin_tr (time    begin_time=0,
    #                                                integer parent_handle=0)
    def m_begin_tr(self, begin_time=0, parent_handle=0):
        m_begin_tr = 0
        tmp_time = begin_time
        if begin_time == 0:
            tmp_time = sv.realtime()
        parent_recorder = None

        if parent_handle != 0:
            parent_recorder = UVMRecorder.get_recorder_from_handle(parent_handle)

        # If we haven't ended the previous record, end it.
        if self.tr_recorder is not None:
            # Don't free the handle, someone else may be using it...
            self.end_tr(tmp_time)

        # May want to establish predecessor/successor relation
        # (don't free handle until then)
        if self.is_recording_enabled():
            db = self.stream_handle.get_db()
            self.end_time = -1
            self.begin_time = tmp_time

            if (parent_recorder is None):
                self.tr_recorder = self.stream_handle.open_recorder(self.get_type_name(),
                        self.begin_time,
                        "Begin_No_Parent, Link")
            else:
                self.tr_recorder = self.stream_handle.open_recorder(self.get_type_name(),
                        self.begin_time,
                        "Begin_End, Link")

                if (self.tr_recorder is not None):
                    pass
                    #link = uvm_parent_child_link::get_link(parent_recorder, self.tr_recorder)
                    # TODO
                    #db.establish_link(link)

            if (self.tr_recorder is not None):
                m_begin_tr = self.tr_recorder.get_handle()
            else:
                m_begin_tr = 0
        else:
            self.tr_recorder = None
            self.end_time = -1
            self.begin_time = tmp_time
            m_begin_tr = 0

        self.do_begin_tr()  # execute callback before event trigger
        self.begin_event.trigger()
        return m_begin_tr

        #endfunction

    #
    #endclass

#------------------------------------------------------------------------------
# IMPLEMENTATION
#------------------------------------------------------------------------------

# get_initiator
# ------------
#
#function uvm_component uvm_transaction::get_initiator()
#  return initiator
#endfunction
#
#
#
# do_print
# --------
#
#function void uvm_transaction::do_print (uvm_printer printer)
#  string str
#  uvm_component tmp_initiator; //work around $swrite bug
#  super.do_print(printer)
#  if(accept_time != -1)
#    printer.print_time("accept_time", accept_time)
#  if(begin_time != -1)
#    printer.print_time("begin_time", begin_time)
#  if(end_time != -1)
#    printer.print_time("end_time", end_time)
#  if(initiator is not None):
#    tmp_initiator = initiator
#    $swrite(str,"@%0d", tmp_initiator.get_inst_id())
#    printer.print_generic("initiator", initiator.get_type_name(), -1, str)
#  end
#endfunction
#
#function void uvm_transaction::do_copy (uvm_object rhs)
#  uvm_transaction txn
#  super.do_copy(rhs)
#  if(rhs is None) return
#  if(!$cast(txn, rhs) ) return
#
#  accept_time = txn.accept_time
#  begin_time = txn.begin_time
#  end_time = txn.end_time
#  initiator = txn.initiator
#  stream_handle = txn.stream_handle
#  self.tr_recorder = txn.self.tr_recorder
#endfunction
#
# do_record
# ---------
#
#function void uvm_transaction::do_record (uvm_recorder recorder)
#  string s
#  super.do_record(recorder)
#  if(accept_time != -1)
#     recorder.record_field("accept_time", accept_time, $bits(accept_time), UVM_TIME)
#  if(initiator is not None):
#    uvm_recursion_policy_enum p = recorder.policy
#    recorder.policy = UVM_REFERENCE
#    recorder.record_object("initiator", initiator)
#    recorder.policy = p
#  end
#endfunction
#



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import *

from uvm.base.sv import sv_if


class apb_if(sv_if):
    #   wire [31:0] paddr
    #   wire        psel
    #   wire        penable
    #   wire        pwrite
    #   wire [31:0] prdata
    #   wire [31:0] pwdata

    def __init__(self, dut, bus_map=None, name="apb"):
        if bus_map is None:
            bus_map = {"clk": "pclk", "psel": "psel", "penable": "penable",
                    "pwrite": "pwrite", "prdata": "prdata", "pwdata": "pwdata",
                    "paddr": "paddr"}
        super().__init__(dut, name, bus_map)
        self.rst = dut.rst


    
    async def start(self):
        await Timer(0)

    #   clocking mck @(posedge pclk)
    #      output paddr, psel, penable, pwrite, pwdata
    #      input  prdata
    #
    #      sequence at_posedge
    #         1
    #      endsequence : at_posedge
    #   endclocking: mck
    #
    #   clocking sck @(posedge pclk)
    #      input  paddr, psel, penable, pwrite, pwdata
    #      output prdata
    #
    #      sequence at_posedge_; // FIXME todo review
    #         1
    #      endsequence : at_posedge_
    #   endclocking: sck
    #
    #   clocking pck @(posedge pclk)
    #      input paddr, psel, penable, pwrite, prdata, pwdata
    #   endclocking: pck
    #
    #   modport master(clocking mck)
    #   modport slave(clocking sck)
    #   modport passive(clocking pck)
    #
    #endinterface: apb_if
    #
    #`endif



#
#------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http:#www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------


from ..base.uvm_component import UVMComponent

class UVMTest(UVMComponent):
    """
    This class is the virtual base class for the user-defined tests.

    The `UVMTest` class should be used as the base class for user-defined
    tests. Doing so provides the ability to select which test to execute using
    the UVM_TESTNAME command line or argument to the `UVMRoot.run_test` task.

    For example

    .. code-block:: shell

      prompt> SIM_COMMAND +UVM_TESTNAME=test_bus_retry

    The global run_test() task should be specified inside a function decorated
    with `@cocotb.test()` (the parentheses are important!) such as

    .. code-block:: python
        import cocotb
        from uvm import run_test
        @cocotb.test()
        def initial(dut):
            await run_test();

    Multiple tests, identified by their type name, can be
    selected for execution from the command line.
    Random seed selection is also available on the command line.

    If +UVM_TESTNAME=test_name is specified, then an object of type 'test_name'
    is created by factory and phasing begins. Here, it is presumed that the
    test will instantiate the test environment, or the test environment will
    have already been instantiated before the call to `run_test()`.

    If the specified test_name cannot be created by the `UVMFactory`, then a
    fatal error occurs. If run_test() is called without UVM_TESTNAME being
    specified, then all components constructed before the call to run_test will
    be cycled through their simulation phases.

    Deriving from `UVMTest` will allow you to distinguish tests from other
    component types that inherit from `UVMComponent` directly. Such tests will
    automatically inherit features that may be added to uvm_test in the future.
    """

    type_name = "uvm_test"

    # Function: new
    #
    # Creates and initializes an instance of this class using the normal
    # constructor arguments for <uvm_component>: ~name~ is the name of the
    # instance, and ~parent~ is the handle to the hierarchical parent, if any.
    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)

    def get_type_name(self):
      return UVMTest.type_name




import unittest
import re
from uvm.base.uvm_printer import (UVMPrinter, UVMLinePrinter, UVMTreePrinter,
    UVMTablePrinter, UVMJSONPrinter)

from uvm.base.uvm_object import UVMObject
from uvm.base.uvm_object_globals import UVM_DEC


class TestUVMPrinter(unittest.TestCase):

    def test_printer_init(self):
        printer = UVMPrinter()

    def test_print_array_header(self):
        printer = UVMPrinter()
        printer.print_array_header("xxx", 10)
        row_info = printer.m_rows[0]


class TestUVMTablePrinter(unittest.TestCase):

    def test_print_field_int(self):
        tprinter = UVMTablePrinter()
        tprinter.print_field_int("my_number", 0x5678, 32)
        result = tprinter.emit()
        self.assertRegex(result, 'number.*5678')

    def test_table_printer(self):
        tprinter = UVMTablePrinter()
        tprinter.knobs.default_radix = UVM_DEC
        tprinter.print_field_int("number", 1234, 32)
        tprinter.print_string("my_string", "xxx yyy")
        tprinter.print_real("my_real", 1.2345)
        result = tprinter.emit()
        self.assertRegex(result, 'number.*1234')
        self.assertRegex(result, 'my_string.*xxx yyy')
        self.assertRegex(result, r'my_real.*real.*1\.2345')


class TestUVMLinePrinter(unittest.TestCase):

    def test_line_printer(self):
        lprinter = UVMLinePrinter()
        lprinter.knobs.default_radix = UVM_DEC
        lprinter.print_field_int("number", 1234, 32)
        lprinter.print_string("my_string", "xxx yyy")
        lprinter.print_real("my_real", 1.2345)
        obj = UVMObject("test obj123")
        lprinter.print_object(obj.get_name(), obj, ".")
        result = lprinter.emit()
        self.assertRegex(result, 'number.*1234')
        self.assertRegex(result, 'my_string.*xxx yyy')
        self.assertRegex(result, 'my_real: 1.2345')
        self.assertRegex(result, 'test obj123')


if __name__ == '__main__':
    unittest.main()




import unittest
from uvm.base.uvm_resource import UVMResource, UVMResourcePool


class TestUVMResource(unittest.TestCase):

    def test_scope_match(self):
        nname1 = "my_comp.field1"
        nname2 = "my_comp.*.field1"
        rr = UVMResource("fname", "my_comp.field1")
        self.assertEqual(True, rr.match_scope(nname1))
        self.assertEqual(False, rr.match_scope(nname2))


    def test_write_read(self):
        rsc = UVMResource("rname", "scope_name")
        rsc.write(123, accessor=None)
        val = rsc.read()
        self.assertEqual(val, 123)


class TestUVMResourcePool(unittest.TestCase):

    def test_get(self):
        pool = UVMResourcePool.get()

    def test_regex_lookup(self):
        # pool = UVMPool()
        field_name = "field"
        inst_re_name = "globbed_name.*"
        r = UVMResource(field_name, inst_re_name)
        r.write(666)
        # lookup = inst_re_name + "__M_UVM__" + field_name
        # pool.add(lookup, r)
        inst_name = "globbed_name.child"
        rp = UVMResourcePool.get()
        rp.set(r)
        rq = rp.lookup_regex_names(inst_name, field_name)
        self.assertEqual(len(rq), 1)

        # Add another resource and make sure we're not getting it
        r2 = UVMResource('field_xxx', 'inst_yyy')
        r2.write(444)
        rp.set(r2)
        rq = rp.lookup_regex_names(inst_name, field_name)
        self.assertEqual(len(rq), 1)

        nname1 = "my_comp.field1"
        nname2 = "my_comp.*.field1"
        rn1 = UVMResource(field_name, nname1)
        rn1.write(111)
        rn2 = UVMResource(field_name, nname2)
        rn2.write(222)
        rp.set(rn1)
        rp.set(rn2)
        rq = rp.lookup_regex_names("my_comp.field1", field_name)
        self.assertEqual(rq[0].read(), 111)


    def test_set_get(self):
        rsrc = UVMResource("name_xxx", "scope_yyy")
        rsrc.write(567, accessor=None)
        pool = UVMResourcePool.get()
        pool.set(rsrc, False)
        rq = pool.lookup_name("scope_yyy", "name_xxx")
        self.assertEqual(len(rq), 1)
        self.assertEqual(rq[0].read(), 567)


if __name__ == '__main__':
    unittest.main()




class RandomizeError(Exception):
    """
    Exception which is thrown when constrained randomization
    fails
    """
    pass


class TLMError(Exception):
    """
    Exception which is thrown when nonblocking TLM method fails
    fails
    """
    pass



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc. 
#//   Copyright 2010 Synopsys, Inc.
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#
#//------------------------------------------------------------------------------
#//
#// MACROS for uvm_printer usage
#//
#// Provides a set of printing macros that will call appropriate print methods
#// inside of a uvm_printer object. All macros have two versions: one assumes
#// a printer named printer is available in scope; the other takes a printer
#// argument.
#//
#//------------------------------------------------------------------------------
#
#`ifndef UVM_PRINTER_DEFINES_SVH
#`define UVM_PRINTER_DEFINES_SVH
#
#// uvm_print_int*
#// --------------
#
#`define uvm_print_int(F, R) \
#  `uvm_print_int3(F, R, uvm_default_printer)
#
#`define uvm_print_int3(F, R, P) \
#   do begin \
#     uvm_printer p__; \
#     if(P!=null) p__ = P; \
#     else p__ = uvm_default_printer; \
#     `uvm_print_int4(F, R, `"F`", p__) \
#   end while(0);
#
#`define uvm_print_int4(F, R, NM, P) \
#    if ($bits(F) > 64) \
#      P.print_field(NM, F, $bits(F), R, "["); \
#    else \
#      P.print_field_int(NM, F, $bits(F), R, "["); 
#
#
#// uvm_print_enum
#// --------------
#
#`define uvm_print_enum(T, F, NM, P) \
#    P.print_generic(NM, `"T`", $bits(F), F.name(), "[");
#
#
#// uvm_print_object*
#// -----------------
#
#`define uvm_print_object(F) \
#  `uvm_print_object2(F, uvm_default_printer)
#
#`define uvm_print_object2(F, P) \
#   do begin \
#     uvm_printer p__; \
#     if(P!=null) p__ = P; \
#     else p__ = uvm_default_printer; \
#     p__.print_object(`"F`", F, "["); \
#   end while(0);
#
#
#// uvm_print_string*
#// -----------------
#
#`define uvm_print_string(F) \
#  `uvm_print_string2(F, uvm_default_printer)
#
#`define uvm_print_string2(F, P) \
#   do begin \
#     uvm_printer p__; \
#     if(P!=null) p__ = P; \
#     else p__ = uvm_default_printer; \
#     p__.print_string(`"F`", F, "["); \
#   end while(0);
#
#
#// uvm_print_array*
#// ----------------
#
#`define uvm_print_array_int(F, R) \
#  `uvm_print_array_int3(F, R, uvm_default_printer)
#   
#`define uvm_print_array_int3(F, R, P) \
#  `uvm_print_qda_int4(F, R, P, da)
#
#
#// uvm_print_sarray*
#// -----------------
#
#`define uvm_print_sarray_int3(F, R, P) \
#  `uvm_print_qda_int4(F, R, P, sa)
#
#`define uvm_print_qda_int4(F, R, P, T) \
#  begin \
#    uvm_printer p__; \
#    uvm_printer_knobs k__; \
#    int curr, max__; max__=0; curr=0; \
#    if(P!=null) p__ = P; \
#    else p__ = uvm_default_printer; \
#    max__ = $right(F)+1; \
#//    max__=$size(F); \
#    p__.print_array_header (`"F`", max__,`"T``(integral)`"); \
#    k__ = p__.knobs; \
#    if((p__.knobs.depth == -1) || (p__.m_scope.depth() < p__.knobs.depth+1)) \
#    begin \
#      foreach(F[i__]) begin \
#        if(k__.begin_elements == -1 || k__.end_elements == -1 || curr < k__.begin_elements ) begin \
#          `uvm_print_int4(F[curr], R, p__.index_string(curr), p__) \
#        end \
#        else break; \
#        curr++; \
#      end \
#      if(curr<max__) begin \
#        if((max__-k__.end_elements) > curr) curr = max__-k__.end_elements; \
#        if(curr<k__.begin_elements) curr = k__.begin_elements; \
#        else begin \
#          p__.print_array_range(k__.begin_elements, curr-1); \
#        end \
#        for(curr=curr; curr<max__; ++curr) begin \
#          `uvm_print_int4(F[curr], R, p__.index_string(curr), p__) \
#        end \
#      end \
#    end \
#    p__.print_array_footer(max__); \
#    //p__.print_footer(); \
#  end
# 
#`define uvm_print_qda_enum(F, P, T, ET) \
#  begin \
#    uvm_printer p__; \
#    uvm_printer_knobs k__; \
#    int curr, max__; max__=0; curr=0; \
#    if(P!=null) p__ = P; \
#    else p__ = uvm_default_printer; \
#    foreach(F[i]) max__ = i+1; \
#    //max__=$size(F); \
#    p__.print_array_header (`"F`", max__,`"T``(``ET``)`"); \
#    k__ = p__.knobs; \
#    if((p__.knobs.depth == -1) || (p__.m_scope.depth() < p__.knobs.depth+1)) \
#    begin \
#      foreach(F[i__]) begin \
#        if(k__.begin_elements == -1 || k__.end_elements == -1 || curr < k__.begin_elements ) begin \
#          `uvm_print_enum(ET, F[curr], p__.index_string(curr), p__) \
#        end \
#        else break; \
#        curr++; \
#      end \
#      if(curr<max__) begin \
#        if((max__-k__.end_elements) > curr) curr = max__-k__.end_elements; \
#        if(curr<k__.begin_elements) curr = k__.begin_elements; \
#        else begin \
#          p__.print_array_range(k__.begin_elements, curr-1); \
#        end \
#        for(curr=curr; curr<max__; ++curr) begin \
#          `uvm_print_enum(ET, F[curr], p__.index_string(curr), p__) \
#        end \
#      end \
#    end \
#    p__.print_array_footer(max__); \
#    //p__.print_footer(); \
#  end
# 
#`define uvm_print_queue_int(F, R) \
#  `uvm_print_queue_int3(F, R, uvm_default_printer)
#
#`define uvm_print_queue_int3(F, R, P) \
#  `uvm_print_qda_int3(F, R, P, queue)
#
#`define uvm_print_array_object(F,FLAG) \
#  `uvm_print_array_object3(F, uvm_default_printer,FLAG)
#   
#`define uvm_print_sarray_object(F,FLAG) \
#  `uvm_print_sarray_object3(F, uvm_default_printer,FLAG)
#   
#`define uvm_print_array_object3(F, P,FLAG) \
#  `uvm_print_object_qda4(F, P, da,FLAG)
#
#`define uvm_print_sarray_object3(F, P,FLAG) \
#  `uvm_print_object_qda4(F, P, sa,FLAG)
#
#`define uvm_print_object_qda4(F, P, T,FLAG) \
#  do begin \
#    int curr, max__; \
#    uvm_printer p__; \
#    max__=0; curr=0; \
#    if(P!=null) p__ = P; \
#    else p__ = uvm_default_printer; \
#    //max__=$size(F); \
#    foreach(F[i]) max__ = i+1; \
#\
#    //p__.print_header();\
#\
#    p__.m_scope.set_arg(`"F`");\
#    p__.print_array_header(`"F`", max__, `"T``(object)`");\
#    if((p__.knobs.depth == -1) || (p__.knobs.depth+1 > p__.m_scope.depth())) \
#    begin\
#      for(curr=0; curr<max__ && (p__.knobs.begin_elements == -1 || \
#         p__.knobs.end_elements == -1 || curr<p__.knobs.begin_elements); ++curr) begin \
#        if(((FLAG)&UVM_REFERENCE) == 0) \
#          p__.print_object(p__.index_string(curr), F[curr], "[");\
#        else \
#          p__.print_object_header(p__.index_string(curr), F[curr], "[");\
#      end \
#      if(curr<max__) begin\
#        curr = max__-p__.knobs.end_elements;\
#        if(curr<p__.knobs.begin_elements) curr = p__.knobs.begin_elements;\
#        else begin\
#          p__.print_array_range(p__.knobs.begin_elements, curr-1);\
#        end\
#        for(curr=curr; curr<max__; ++curr) begin\
#          if(((FLAG)&UVM_REFERENCE) == 0) \
#            p__.print_object(p__.index_string(curr), F[curr], "[");\
#          else \
#            p__.print_object_header(p__.index_string(curr), F[curr], "[");\
#        end \
#      end\
#    end \
#\
#    p__.print_array_footer(max__); \
#    //p__.print_footer(); \
#  end while(0);
# 
#`define uvm_print_object_queue(F,FLAG) \
#  `uvm_print_object_queue3(F, uvm_default_printer,FLAG)
#   
#`define uvm_print_object_queue3(F, P,FLAG) \
#  do begin \
#    `uvm_print_object_qda4(F,P, queue,FLAG); \
#  end while(0);
# 
#`define uvm_print_array_string(F) \
#  `uvm_print_array_string2(F, uvm_default_printer)
#   
#`define uvm_print_array_string2(F, P) \
#   `uvm_print_string_qda3(F, P, da)
#
#`define uvm_print_sarray_string2(F, P) \
#   `uvm_print_string_qda3(F, P, sa)
#
#`define uvm_print_string_qda3(F, P, T) \
#  do begin \
#    int curr, max__; \
#    uvm_printer p__; \
#    max__=0; curr=0; \
#    //max__=$size(F); \
#    foreach(F[i]) max__ = i+1; \
#    if(P!=null) p__ = P; \
#    else p__ = uvm_default_printer; \
#\
#    //p__.print_header();\
#\
#    p__.m_scope.set_arg(`"F`");\
#    p__.print_array_header(`"F`", max__, `"T``(string)`");\
#    if((p__.knobs.depth == -1) || (p__.knobs.depth+1 > p__.m_scope.depth())) \
#    begin\
#      for(curr=0; curr<max__ && curr<p__.knobs.begin_elements; ++curr) begin\
#        p__.print_string(p__.index_string(curr), F[curr], "[");\
#      end \
#      if(curr<max__) begin\
#        curr = max__-p__.knobs.end_elements;\
#        if(curr<p__.knobs.begin_elements) curr = p__.knobs.begin_elements;\
#        else begin\
#          p__.print_array_range(p__.knobs.begin_elements, curr-1);\
#        end\
#        for(curr=curr; curr<max__; ++curr) begin\
#          p__.print_string(p__.index_string(curr), F[curr], "[");\
#        end \
#      end\
#    end \
#\
#    p__.print_array_footer(max__); \
#    //p__.print_footer(); \
#  end while(0);
# 
#`define uvm_print_string_queue(F) \
#  `uvm_print_string_queue2(F, uvm_default_printer)
#   
#`define uvm_print_string_queue2(F, P) \
#  do begin \
#    `uvm_print_string_qda3(F,P, queue); \
#  end while(0);
#
#//-----------------------------------------------------------------------------
#//
#// Associative array printing methods
#//
#//-----------------------------------------------------------------------------
#`define uvm_print_aa_string_int(F) \
#  `uvm_print_aa_string_int3(F, R, uvm_default_printer)
#
#
#`define uvm_print_aa_string_int3(F, R, P) \
#  begin \
#    uvm_printer p__; \
#    uvm_printer_knobs k__; \
#    if(P!=null) p__ = P; \
#    else p__ = uvm_default_printer; \
#    p__.print_array_header (`"F`", F.num(), "aa(int,string)"); \
#    k__ = p__.knobs; \
#    if((p__.knobs.depth == -1) || (p__.m_scope.depth() < p__.knobs.depth+1)) \
#    begin \
#      foreach(F[string_aa_key]) \
#          `uvm_print_int4(F[string_aa_key], R,  \
#                                {"[", string_aa_key, "]"}, p__) \
#    end \
#    p__.print_array_footer(F.num()); \
#    //p__.print_footer(); \
#  end
#
#`define uvm_print_aa_string_object(F,FLAG) \
#  `uvm_print_aa_string_object_3(F, uvm_default_printer,FLAG)
#  
#`define uvm_print_aa_string_object3(F, P,FLAG) \
#  begin \
#    uvm_printer p__; \
#    uvm_printer_knobs k__; \
#    uvm_object u__; \
#    if(P!=null) p__ = P; \
#    else p__ = uvm_default_printer; \
#    p__.print_array_header (`"F`", F.num(), "aa(object,string)"); \
#    k__ = p__.knobs; \
#    if((p__.knobs.depth == -1) || (p__.m_scope.depth() < p__.knobs.depth+1)) \
#    begin \
#      foreach(F[string_aa_key]) begin \
#          if(((FLAG)&UVM_REFERENCE)==0) \
#            p__.print_object({"[", string_aa_key, "]"}, F[string_aa_key], "[");\
#          else \
#            p__.print_object_header({"[", string_aa_key, "]"}, F[string_aa_key], "[");\
#      end \
#    end \
#    p__.print_array_footer(F.num()); \
#    //p__.print_footer(); \
#  end
#
#`define uvm_print_aa_string_string(F) \
#  `uvm_print_aa_string_string_2(F, uvm_default_printer)
#  
#`define uvm_print_aa_string_string2(F, P) \
#  begin \
#    uvm_printer p__; \
#    uvm_printer_knobs k__; \
#    if(P!=null) p__ = P; \
#    else p__ = uvm_default_printer; \
#    p__.print_array_header (`"F`", F.num(), "aa(string,string)"); \
#    k__ = p__.knobs; \
#    if((p__.knobs.depth == -1) || (p__.m_scope.depth() < p__.knobs.depth+1)) \
#    begin \
#      foreach(F[string_aa_key]) \
#          p__.print_string({"[", string_aa_key, "]"}, F[string_aa_key], "["); \
#    end \
#    p__.print_array_footer(F.num()); \
#    //p__.print_footer(); \
#  end
#
#`define uvm_print_aa_int_object(F,FLAG) \
#  `uvm_print_aa_int_object_3(F, uvm_default_printer,FLAG)
#
#`define uvm_print_aa_int_object3(F, P,FLAG) \
#  begin \
#    uvm_printer p__; \
#    uvm_printer_knobs k__; \
#    uvm_object u__; \
#    int key; \
#    if(P!=null) p__ = P; \
#    else p__ = uvm_default_printer; \
#    p__.print_array_header (`"F`", F.num(), "aa(object,int)"); \
#    k__ = p__.knobs; \
#    if((p__.knobs.depth == -1) || (p__.m_scope.depth() < p__.knobs.depth+1)) \
#    begin \
#      foreach(F[key]) begin \
#          $swrite(__m_uvm_status_container.stringv, "[%0d]", key); \
#          if(((FLAG)&UVM_REFERENCE)==0) \
#            p__.print_object(__m_uvm_status_container.stringv, F[key], "[");\
#          else \
#            p__.print_object_header(__m_uvm_status_container.stringv, F[key], "[");\
#      end \
#    end \
#    p__.print_array_footer(F.num()); \
#    //p__.print_footer(); \
#  end
#
#`define uvm_print_aa_int_key4(KEY, F, R, P) \
#  begin \
#    uvm_printer p__; \
#    uvm_printer_knobs k__; \
#    if(P!=null) p__ = P; \
#    else p__ = uvm_default_printer; \
#    __m_uvm_status_container.stringv = "aa(int,int)"; \
#    for(int i=0; i<__m_uvm_status_container.stringv.len(); ++i) \
#      if(__m_uvm_status_container.stringv[i] == " ") \
#        __m_uvm_status_container.stringv[i] = "_"; \
#    p__.print_array_header (`"F`", F.num(), __m_uvm_status_container.stringv); \
#    k__ = p__.knobs; \
#    if((p__.knobs.depth == -1) || (p__.m_scope.depth() < p__.knobs.depth+1)) \
#    begin \
#      foreach(F[aa_key]) begin \
#          `uvm_print_int4(F[aa_key], R,  \
#                                {"[", $sformatf("%0d",aa_key), "]"}, p__) \
#      end \
#    end \
#    p__.print_array_footer(F.num()); \
#    //p__.print_footer(); \
#  end
#
#`endif




import unittest

from uvm.base.uvm_globals import (
    uvm_report_enabled,
    uvm_is_match
)

from uvm.base.uvm_object_globals import (
    UVM_INFO, UVM_MEDIUM, UVM_HIGH,
    UVM_FATAL
)


class TestUVMGlobals(unittest.TestCase):

    def test_uvm_report_enabled(self):
        # Default verbosity that is set
        self.assertTrue(uvm_report_enabled(UVM_MEDIUM, UVM_INFO))
        self.assertFalse(uvm_report_enabled(UVM_HIGH, UVM_INFO))
        self.assertTrue(uvm_report_enabled(UVM_MEDIUM, UVM_FATAL))


    def test_uvm_is_match(self):
        self.assertTrue(uvm_is_match("my_name*", "my_name.child1.c2"))
        self.assertFalse(uvm_is_match("your_name*", "my_name.child1.c2"))
        self.assertTrue(uvm_is_match("my_name????", "my_name.abc"))
        self.assertFalse(uvm_is_match("my_name??????", "my_name.abc"))
        self.assertTrue(uvm_is_match("zzz*www??yy", "zzz_abcdefg_wwwKKyy"))


if __name__ == '__main__':
    unittest.main()




# UNIT TESTS

import unittest
from uvm.base.uvm_root import UVMRoot


class TestUVMRoot(unittest.TestCase):

    def test_name(self):
        root = UVMRoot()
        self.assertEqual(root.get_name(), "")

    def test_singleton(self):
        root1 = UVMRoot.m_uvm_get_root()
        root2 = UVMRoot.m_uvm_get_root()
        self.assertEqual(root1, root2)

    async def test_run_phase(self):
        root = UVMRoot()
        await root.run_phase()


if __name__ == '__main__':
    unittest.main()




import unittest
# import re
from uvm.base.uvm_packer import (UVMPacker, MASK_INT)
from uvm.base.uvm_object import UVMObject
from uvm.base.sv import sv


class TestUVMPacker(unittest.TestCase):

    def test_flip_bit_order(self):
        packer = UVMPacker()
        bin1 = 0b0010
        flip1 = packer.flip_bit_order(bin1, 8)
        self.assertEqual(flip1, 0b01000000)
        self.assertEqual(packer.flip_bit_order(flip1, 8), bin1)
        for _ in range(16):
            sizes = [8, 16, 24, 32, sv.urandom_range(3, 64)]
            for size in sizes:
                value = sv.urandom_range(1, (1 << size) - 1)
                flip1 = packer.flip_bit_order(value, size)
                double_flip1 = packer.flip_bit_order(flip1, size)
                self.assertEqual(double_flip1, value)

    def test_pack_int(self):
        packer = UVMPacker()
        packer.big_endian = 0
        myshort = 0x1234
        mybyte = 0x67
        packer.pack_field_int(myshort, 16)
        packer.pack_field_int(mybyte, 8)

        packer.set_packed_size()
        packed_size = packer.get_packed_size()
        self.assertEqual(packed_size, 24)

        bits = packer.get_packed_bits()
        self.assertEqual(bits, (0x67 << 16 | 0x1234))
        bit0 = packer.get_bit(0)
        self.assertEqual(bit0, 0x4 & 0x1)
        self.assertEqual(packer.get_bit(2), (0x4 >> 2)  & 0x1)

    def test_pack_unpack_bytes(self):
        packer = UVMPacker()
        packer.big_endian = 0
        arr_bytes = [0x01, 0x03, 0x07, 0xAB]
        packer.pack_bytes(arr_bytes, 32)
        packer.set_packed_size()
        bits = packer.get_packed_bits()
        self.assertEqual(bits, 0xAB070301)
        arr = [0] * 4
        new_bytes = packer.unpack_bytes(arr)
        self.assertEqual(new_bytes, arr_bytes)
        my_bytes = packer.get_bytes()
        self.assertEqual(my_bytes, arr_bytes)


    def test_pack_unpack_ints(self):
        packer = UVMPacker()
        packer.big_endian = 0
        arr_int = [0x0123, 0x4567, 0x0089, 0x00AB]
        packer.pack_ints(arr_int, 4 * 32)
        packer.set_packed_size()
        bits = packer.get_packed_bits()
        self.assertEqual(bits, 0x000000AB000000890000456700000123)
        arr = [0] * 4
        new_ints = packer.unpack_ints(arr)
        self.assertEqual(new_ints, arr_int)

        my_ints = packer.get_ints()
        self.assertEqual(my_ints, arr_int)

        packer.reset()
        packer.pack_ints(arr_int, 4 * 32)
        packer.set_packed_size()
        self.assertEqual(packer.unpack_field_int(128), 0x000000AB000000890000456700000123)

        packer.reset()
        packer.pack_ints(arr_int, 4 * 32)
        packer.set_packed_size()
        self.assertEqual(packer.unpack_field_int(64), 0x0000456700000123)

        packer.reset()
        packer.pack_ints(arr_int, 4 * 32)
        packer.set_packed_size()
        self.assertEqual(packer.unpack_field_int(32), 0x00000123)


    def test_unpack_field(self):
        num_ints = int(4096 / 32)
        arr_ints = []
        for i in range(num_ints):
            arr_ints.append(1 + i**2)
        packer = UVMPacker()
        packer.big_endian = 0
        packer.pack_ints(arr_ints, num_ints * 32)
        packer.set_packed_size()
        val_4096b = packer.unpack_field(4096)
        # Verify the results
        for i in range(num_ints):
            value = (val_4096b >> (32*i)) & MASK_INT
            self.assertEqual(value, arr_ints[i])

    def test_pack_unpack_string(self):
        str_list = ["Otatko työ mämmiä, Åke?", 'abcd_efgh_1234_6789']
        for test_str in str_list:
            packer = UVMPacker()
            packer.big_endian = 0
            packer.pack_string(test_str)
            packer.set_packed_size()
            new_str = packer.unpack_string()
            self.assertEqual(new_str, test_str)

    def test_big_endian(self):
        for be in [0, 1]:
            packer = UVMPacker()
            packer.big_endian = be
            packer.pack_field_int(0x123, 32)
            packer.pack_field_int(0xABC, 32)
            packer.pack_field_int(0xFACE, 64)
            packer.pack_field_int(0x77, 8)
            packer.set_packed_size()
            val1 = packer.unpack_field(32)
            val2 = packer.unpack_field(32)
            val3 = packer.unpack_field(64)
            val4 = packer.unpack_field(8)
            self.assertEqual(hex(val1), hex(0x123))
            self.assertEqual(hex(val2), hex(0xABC))
            self.assertEqual(hex(val3), hex(0xFACE))
            self.assertEqual(hex(val4), hex(0x77))




if __name__ == '__main__':
    unittest.main()



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------


#  // This simple example shows how the uvm_test_done objection is
#  // used to coordinate end of test activity. For an example of
#  // using end of test coordination in the context of a full
#  // environment, refer to the xbus example.
#
#  // In this example, a drain time of 10 is set on the component.
#  // The component then forks 4 processes which consume different
#  // amounts of time. When the last process is done (time 50),
#  // the drain time takes effect. The test is completed at time
#  // 60.
#
#  // The example also shows usage of the component objection
#  // callbacks. In this case, the dropped callback is used, but
#  // the raised and all_dropped work similarly (except the
#  // all dropped is a time-consuming task).

import cocotb
from cocotb.triggers import Timer

from uvm.comps.uvm_test import *
from uvm.macros import *
from uvm import *


class simple_test(UVMTest):
    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.sem = UVMMailbox(1)
        self.sem.try_put(1)
        #self.sem.debug_enabled = True
        self.proc_finished = {}
        self.largest_delay = 50


    async def run_phase(self, phase):
        # Set a drain time on the objection if needed
        uvm_info("drain", "Setting drain time of 10", UVM_NONE)

        # DEPRECATED
        #uvm_test_done.set_drain_time(self, 10)

        # Run a bunch of processes in parallel
        #fork
        p0 = cocotb.fork(self.doit(35, phase, 0))
        p1 = cocotb.fork(self.doit(25, phase, 1))
        p2 = cocotb.fork(self.doit(self.largest_delay, phase, 2))
        p3 = cocotb.fork(self.doit(15, phase, 3))
        p4 = cocotb.fork(self.do_without_object(self.largest_delay + 5, 4))

        uvm_info("all_procs_forked", "Proceeding to await proc4", UVM_NONE)

        # p4 runs longest, but does not raise objection, so simulation should
        # terminate at largest_delay, not when p4 finishes (+5)
        await p4
        #yield [p0.join(), p1.join(), p2.join(), p3.join(), ]
        #join

    s_inst = 0
    # A simple task that consumes some time.

    async def doit(self, delay, phase, _id):
        arr = []
        # Raise an objection before starting the activity
        phase.raise_objection(self)
        await self.sem.get(arr)
        inst = simple_test.s_inst
        simple_test.s_inst += 1
        await self.sem.put(1)

        uvm_info("doit", sv.sformatf("Starting doit (%0d) with delay %0t",
            inst, delay), UVM_NONE)
        await Timer(delay, "NS")
        uvm_info("doit", sv.sformatf("Ending doit (%0d)", inst), UVM_NONE)

        # Drop the objection when done
        arr = []
        await self.sem.get(arr)
        self.proc_finished[_id] = True
        await self.sem.put(1)
        phase.drop_objection(self)
        uvm_info("doit", sv.sformatf("Really finished now (%0d)", inst), UVM_NONE)


    async def do_without_object(self, delay, _id):
        await Timer(1, "NS")
        arr = []
        await self.sem.get(arr)
        inst = simple_test.s_inst
        simple_test.s_inst += 1
        await self.sem.put(1)

        uvm_info("doit", sv.sformatf("Starting doit (%0d) with delay %0t",
            inst, delay), UVM_NONE)
        await Timer(delay, "NS")
        uvm_info("doit", sv.sformatf("Ending doit (%0d)", inst), UVM_NONE)

        # Drop the objection when done
        arr = []
        await self.sem.get(arr)
        self.proc_finished[_id] = True
        await self.sem.put(1)

    # Use an objection callback do something when objections are raised or
    # dropped (or all dropped). This example prints some information on each
    # drop.
    def dropped(self, objection, source_obj, description, count):
        uvm_info("dropped",
            sv.sformatf("%d objection(s) dropped from %s, total count is now %0d",
            count, source_obj.get_full_name, objection.get_objection_total(self)),
            UVM_NONE)


    # Check various conditions such as processses finishes and simulation time
    # to verify that objections are working correctly
    def check_phase(self, phase):
        ids = [0, 1, 2, 3]
        for _id, i in enumerate(ids):
            if _id not in self.proc_finished:
                uvm_fatal("PROC_NOT_FINISHED",
                "Inst {} not found from finished procs".format(_id))
            else:
                uvm_info("PROC_OK", "Proc {} finished OK".format(_id), UVM_NONE)

        if 5 in self.proc_finished:
            uvm_fatal("WRONG_PROC_FINISHED", "Proc 5 must not finish")
        finish_time = sv.realtime("NS")
        if finish_time != self.largest_delay:
            uvm_fatal("WRONG_FINISH_TIME", "Exp: {}, Got: {}".format(
                self.largest_delay, finish_time))


# Register with the factory.
uvm_component_utils(simple_test)


@cocotb.test()
async def initial(dut):
    # Run the test
    await run_test("simple_test")



#//----------------------------------------------------------------------
#//   Copyright 2011 Cypress Semiconductor
#//   Copyright 2010 Mentor Graphics Corporation
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

#//----------------------------------------------------------------------
#// Title: UVM Resource Database
#//
#// Topic: Intro
#//
#// The <uvm_resource_db> class provides a convenience interface for
#// the resources facility.  In many cases basic operations such as
#// creating and setting a resource or getting a resource could take
#// multiple lines of code using the interfaces in <uvm_resource_base> or
#// <uvm_resource#(T)>.  The convenience layer in <uvm_resource_db>
#// reduces many of those operations to a single line of code.
#//
#// If the run-time ~+UVM_RESOURCE_DB_TRACE~ command line option is
#// specified, all resource DB accesses (read and write) are displayed.
#//----------------------------------------------------------------------


#//----------------------------------------------------------------------
#// class: uvm_resource_db
#//
#// All of the functions in uvm_resource_db#(T) are static, so they
#// must be called using the :: operator.  For example:
#//
#//|  uvm_resource_db#(int)::set("A", "*", 17, this)
#//
#// The parameter value "int" identifies the resource type as
#// uvm_resource#(int).  Thus, the type of the object in the resource
#// container is int. This maintains the type-safety characteristics of
#// resource operations.
#//
#//----------------------------------------------------------------------
#class uvm_resource_db #(type T=uvm_object)

from .sv import sv
from .uvm_globals import uvm_check_output_args
from ..uvm_macros import uvm_typename
from ..macros.uvm_message_defines import uvm_info
from .uvm_object_globals import UVM_LOW
from .uvm_resource import UVMResource, UVMResourcePool


class UVMResourceDb:

    rsrc_t = UVMResource

    #  typedef uvm_resource #(T) rsrc_t

    def __init__(self, T=None):
        self.T = T

    #  // function: get_by_type
    #  //
    #  // Get a resource by type.  The type is specified in the db
    #  // class parameter so the only argument to this function is the
    #  // ~scope~.
    @classmethod
    def get_by_type(cls, scope):
        return cls.rsrc_t.get_by_type(scope, cls.rsrc_t.get_type())
    #  endfunction

    #  // function: get_by_name
    #  //
    #  // Imports a resource by ~name~.  The first argument is the current
    #  // ~scope~ of the resource to be retrieved and the second argument is
    #  // the ~name~. The ~rpterr~ flag indicates whether or not to generate
    #  // a warning if no matching resource is found.
    #
    #  static function rsrc_t get_by_name(string scope,
    #                                     string name,
    #                                     bit rpterr=1)
    @classmethod
    def get_by_name(cls, scope, name, rpterr=True):
        rsrc = UVMResource.get_by_name(scope, name, rpterr)
        if UVMResourceDbOptions.is_tracing():
            cls.m_show_msg("RSRCDB/GET_BY_NAME", "Resource","get", scope, name,
                    None, rsrc)
        return rsrc

    #  // function: set_default
    #  //
    #  // add a new item into the resources database.  The item will not be
    #  // written to so it will have its default value. The resource is
    #  // created using ~name~ and ~scope~ as the lookup parameters.
    #  static function rsrc_t set_default(string scope, string name)
    @classmethod
    def set_default(cls, scope, name):
        r = UVMResource(name, scope)
        r.set()
        return r

    #  // function- show_msg
    #
    #  // internal helper function to print resource accesses
    #
    #  protected static function void m_show_msg(
    #          input string id,
    #          input string rtype,
    #          input string action,
    #          input string scope,
    #          input string name,
    #          input uvm_object accessor,
    #          input rsrc_t rsrc)
    @classmethod
    def m_show_msg(cls, id, rtype, action, scope, name, accessor, rsrc):
        foo = None
        msg = uvm_typename(foo)
        _name = ""
        if name != "":
            _name = "." + name

        acc_name = "<unknown>"
        if accessor is not None:
            acc_name = accessor.get_full_name()

        rsrc_str = "null (failed lookup)"
        if rsrc is not None:
            rsrc_str = rsrc.convert2string()

        msg = sv.sformatf("%s '%s%s' (type %s) %s by %s = %s",
            rtype, scope, _name, msg, action, acc_name, rsrc_str)
        uvm_info(id, msg, UVM_LOW)


    #  // function: set
    #  //
    #  // Create a new resource, write a ~val~ to it, and set it into the
    #  // database using ~name~ and ~scope~ as the lookup parameters. The
    #  // ~accessor~ is used for auditing.
    #  static function void set(input string scope, input string name,
    #                           T val, input uvm_object accessor = null)
    @classmethod
    def set(cls, scope, name, val, accessor=None):
        rsrc = cls.rsrc_t(name, scope)
        rsrc.write(val, accessor)
        rsrc.set()

        if UVMResourceDbOptions.is_tracing():
            cls.m_show_msg("RSRCDB/SET", "Resource","set", scope, name, accessor, rsrc)
        #  endfunction

    #  // function: set_anonymous
    #  //
    #  // Create a new resource, write a ~val~ to it, and set it into the
    #  // database.  The resource has no name and therefore will not be
    #  // entered into the name map. But is does have a ~scope~ for lookup
    #  // purposes. The ~accessor~ is used for auditing.
    #  static function void set_anonymous(input string scope,
    #                                     T val, input uvm_object accessor = null)
    #
    #    rsrc_t rsrc = new("", scope)
    #    rsrc.write(val, accessor)
    #    rsrc.set()
    #
    #    if(UVMResourceDbOptions::is_tracing())
    #      m_show_msg("RSRCDB/SETANON","Resource", "set", scope, "", accessor, rsrc)
    #  endfunction

    #  // function set_override
    #  //
    #  // Create a new resource, write ~val~ to it, and set it into the
    #  // database.  Set it at the beginning of the queue in the type map and
    #  // the name map so that it will be (currently) the highest priority
    #  // resource with the specified name and type.
    #
    #  static function void set_override(input string scope, input string name,
    #                                    T val, uvm_object accessor = null)
    #    rsrc_t rsrc = new(name, scope)
    #    rsrc.write(val, accessor)
    #    rsrc.set_override()
    #
    #    if(UVMResourceDbOptions::is_tracing())
    #      m_show_msg("RSRCDB/SETOVRD", "Resource","set", scope, name, accessor, rsrc)
    #  endfunction



    #  // function set_override_type
    #  //
    #  // Create a new resource, write ~val~ to it, and set it into the
    #  // database.  Set it at the beginning of the queue in the type map so
    #  // that it will be (currently) the highest priority resource with the
    #  // specified type. It will be normal priority (i.e. at the end of the
    #  // queue) in the name map.
    #
    #  static function void set_override_type(input string scope, input string name,
    #                                         T val, uvm_object accessor = null)
    #    rsrc_t rsrc = new(name, scope)
    #    rsrc.write(val, accessor)
    #    rsrc.set_override(uvm_resource_types::TYPE_OVERRIDE)
    #
    #    if(UVMResourceDbOptions::is_tracing())
    #      m_show_msg("RSRCDB/SETOVRDTYP","Resource", "set", scope, name, accessor, rsrc)
    #  endfunction

    #  // function set_override_name
    #  //
    #  // Create a new resource, write ~val~ to it, and set it into the
    #  // database.  Set it at the beginning of the queue in the name map so
    #  // that it will be (currently) the highest priority resource with the
    #  // specified name. It will be normal priority (i.e. at the end of the
    #  // queue) in the type map.
    #
    #  static function void set_override_name(input string scope, input string name,
    #                                  T val, uvm_object accessor = null)
    #    rsrc_t rsrc = new(name, scope)
    #    rsrc.write(val, accessor)
    #    rsrc.set_override(uvm_resource_types::NAME_OVERRIDE)
    #
    #    if(UVMResourceDbOptions::is_tracing())
    #      m_show_msg("RSRCDB/SETOVRDNAM","Resource", "set", scope, name, accessor, rsrc)
    #  endfunction

    #  // function: read_by_name
    #  //
    #  // locate a resource by ~name~ and ~scope~ and read its value. The value
    #  // is returned through the output argument ~val~.  The return value is a bit
    #  // that indicates whether or not the read was successful. The ~accessor~
    #  // is used for auditing.
    #  static function bit read_by_name(input string scope,
    #                                   input string name,
    #                                   inout T val, input uvm_object accessor = null)
    @classmethod
    def read_by_name(cls, scope, name, val, accessor=None):
        uvm_check_output_args([val])
        rsrc = cls.get_by_name(scope, name)

        if UVMResourceDbOptions.is_tracing():
            cls.m_show_msg("RSRCDB/RDBYNAM","Resource", "read", scope, name, accessor, rsrc)

        if rsrc is None:
            return False

        val.append(rsrc.read(accessor))
        return True


    #  // function: read_by_type
    #  //
    #  // Read a value by type.  The value is returned through the output
    #  // argument ~val~.  The ~scope~ is used for the lookup. The return
    #  // value is a bit that indicates whether or not the read is successful.
    #  // The ~accessor~ is used for auditing.
    #  static function bit read_by_type(input string scope,
    #                                   inout T val,
    #                                   input uvm_object accessor = null)
    #
    #    rsrc_t rsrc = get_by_type(scope)
    #
    #    if(UVMResourceDbOptions::is_tracing())
    #      m_show_msg("RSRCDB/RDBYTYP", "Resource","read", scope, "", accessor, rsrc)
    #
    #    if(rsrc == null)
    #      return 0
    #
    #    val = rsrc.read(accessor)
    #
    #    return 1
    #
    #  endfunction

    #  // function: write_by_name
    #  //
    #  // write a ~val~ into the resources database.  First, look up the
    #  // resource by ~name~ and ~scope~.  If it is not located then add a new
    #  // resource to the database and then write its value.
    #  //
    #  // Because the ~scope~ is matched to a resource which may be a
    #  // regular expression, and consequently may target other scopes beyond
    #  // the ~scope~ argument. Care must be taken with this function. If
    #  // a <get_by_name> match is found for ~name~ and ~scope~ then ~val~
    #  // will be written to that matching resource and thus may impact
    #  // other scopes which also match the resource.
    #  static function bit write_by_name(input string scope, input string name,
    #                                    input T val, input uvm_object accessor = null)
    @classmethod
    def write_by_name(cls, scope, name, val, accessor=None):
        rsrc = cls.get_by_name(scope, name)
        if (UVMResourceDbOptions.is_tracing()):
            cls.m_show_msg("RSRCDB/WR","Resource", "written", scope, name, accessor, rsrc)

        if rsrc is None:
            return 0
        rsrc.write(val, accessor)
        return 1

    #  // function: write_by_type
    #  //
    #  // write a ~val~ into the resources database.  First, look up the
    #  // resource by type.  If it is not located then add a new resource to
    #  // the database and then write its value.
    #  //
    #  // Because the ~scope~ is matched to a resource which may be a
    #  // regular expression, and consequently may target other scopes beyond
    #  // the ~scope~ argument. Care must be taken with this function. If
    #  // a <get_by_name> match is found for ~name~ and ~scope~ then ~val~
    #  // will be written to that matching resource and thus may impact
    #  // other scopes which also match the resource.
    #  static function bit write_by_type(input string scope,
    #                                    input T val, input uvm_object accessor = null)
    #
    #    rsrc_t rsrc = get_by_type(scope)
    #
    #    if(UVMResourceDbOptions::is_tracing())
    #      m_show_msg("RSRCDB/WRTYP", "Resource","written", scope, "", accessor, rsrc)
    #
    #    if(rsrc == null)
    #      return 0
    #
    #    rsrc.write(val, accessor)
    #
    #    return 1
    #  endfunction

    #  // function: dump
    #  //
    #  // Dump all the resources in the resource pool. This is useful for
    #  // debugging purposes.  This function does not use the parameter T, so
    #  // it will dump the same thing -- the entire database -- no matter the
    #  // value of the parameter.
    #
    @classmethod
    def dump(cls):
        rp = UVMResourcePool.get()
        rp.dump()

    #endclass


#//----------------------------------------------------------------------
#// Class: UVMResourceDbOptions
#//
#// Provides a namespace for managing options for the
#// resources DB facility.  The only thing allowed in this class is static
#// local data members and static functions for manipulating and
#// retrieving the value of the data members.  The static local data
#// members represent options and settings that control the behavior of
#// the resources DB facility.
#
#// Options include:
#//
#//  * tracing:  on/off
#//
#//    The default for tracing is off.
#//
#//----------------------------------------------------------------------


class UVMResourceDbOptions:

    ready = False
    tracing = False

    @classmethod
    def turn_on_tracing(cls):
        """         
          Function: turn_on_tracing
         
          Turn tracing on for the resource database. This causes all
          reads and writes to the database to display information about
          the accesses. Tracing is off by default.
         
          This method is implicitly called by the ~+UVM_RESOURCE_DB_TRACE~.
        """
        if cls.ready is False:
            cls.init()
        cls.tracing = True

    @classmethod
    def turn_off_tracing(cls):
        """         
          Function: turn_off_tracing
         
          Turn tracing off for the resource database.

        """
        if cls.ready is False:
            cls.init()
        cls.tracing = False


    @classmethod
    def is_tracing(cls):
        """         
          Function: is_tracing
         
          Returns 1 if the tracing facility is on and 0 if it is off.
         static function bit is_tracing()
        Returns:
        """
        if cls.ready is False:
            cls.init()
        return cls.tracing
    #  endfunction


    @classmethod
    def init(cls):
        """         
         static local function void init()
        """
        trace_args = []
        from .uvm_cmdline_processor import UVMCmdlineProcessor
        clp = UVMCmdlineProcessor.get_inst()

        if clp.get_arg_matches("+UVM_RESOURCE_DB_TRACE", trace_args):
            cls.tracing = True
        cls.ready = True
    #  endfunction



# Returns new resource DB class declaration specialized with type T
def ResourceDbClassFactory(name, argnames, T=None):
    def __init__(self, **kwargs):
        for key, val in kwargs.items():
            if key not in argnames:
                raise TypeError('Arg {} not valid for {}'.format(
                    key, name))
            else:
                setattr(self, key, val)
        UVMResourceDb.__init__(self, T)
    newclass = type(name, (UVMResourceDb,),
        {"__init__": __init__})
    return newclass



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2008 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//
"""
Title: Bit Bashing Test Sequences

This section defines classes that test individual bits of the registers
defined in a register model.
"""

import cocotb
from cocotb.triggers import Timer

from ..uvm_reg_sequence import UVMRegSequence
from ..uvm_reg_model import (UVM_NO_HIER, UVM_NO_CHECK,
        UVM_FRONTDOOR, UVM_IS_OK)
from ...macros import (uvm_object_utils, uvm_error, uvm_info, UVM_REG_DATA_WIDTH,
    uvm_warning)
from ...base import UVMResourceDb, sv, UVM_LOW, UVM_HIGH, uvm_zero_delay


#//----------------------------------------------------------------------------
#// Class: UVMRegSingleBitBashSeq
#//
#// Verify the implementation of a single register
#// by attempting to write 1's and 0's to every bit in it,
#// via every address map in which the register is mapped,
#// making sure that the resulting value matches the mirrored value.
#//
#// If bit-type resource named
#// "NO_REG_TESTS" or "NO_REG_BIT_BASH_TEST"
#// in the "REG::" namespace
#// matches the full name of the register,
#// the register is not tested.
#//
#//| uvm_resource_db#(bit)::set({"REG::",regmodel.blk.r0.get_full_name()},
#//|                            "NO_REG_TESTS", 1, this);
#//
#// Registers that contain fields with unknown access policies
#// cannot be tested.
#//
#// The DUT should be idle and not modify any register during this test.
#//
#//----------------------------------------------------------------------------


class UVMRegSingleBitBashSeq(UVMRegSequence):  # (uvm_sequence #(uvm_reg_item))

    #   // Variable: rg
    #   // The register to be tested
    #   uvm_reg rg

    def __init__(self, name="UVMRegSingleBitBashSeq"):
        super().__init__(name)
        self.rg = None


    async def body(self):
        fields = []  # uvm_reg_field[$]
        mode = [""] * UVM_REG_DATA_WIDTH  # string [`UVM_REG_DATA_WIDTH]
        maps = []  # uvm_reg_map [$]
        dc_mask = [0] * UVM_REG_DATA_WIDTH
        reset_val = 0x0
        n_bits = 0
        field_access = ""
        rg = self.rg

        if self.rg is None:
            uvm_error("UVMRegBitBashSeq", "No register specified to run sequence on")
            return

        # Registers with some attributes are not to be tested
        if (UVMResourceDb.get_by_name("REG::" + rg.get_full_name(),
            "NO_REG_TESTS", 0) is not None or
                  UVMResourceDb.get_by_name("REG::" + rg.get_full_name(),
                      "NO_REG_BIT_BASH_TEST", 0) is not None):
            return

        n_bits = self.rg.get_n_bytes() * 8

        # Let's see what kind of bits we have...
        rg.get_fields(fields)

        # Registers may be accessible from multiple physical interfaces (maps)
        rg.get_maps(maps)

        # TODO Bash the bits in the register via each map
        for j in range(len(maps)):
            status = 0
            val = 0x0
            exp = 0x0
            v = 0x0
            next_lsb = 0
            dc_mask  = [0] * UVM_REG_DATA_WIDTH

            for k in range(len(fields)):

                field_access = fields[k].get_access(maps[j])
                dc = (fields[k].get_compare() == UVM_NO_CHECK)
                lsb = fields[k].get_lsb_pos()
                w   = fields[k].get_n_bits()

                # Ignore Write-only fields because
                # you are not supposed to read them
                if field_access in ["WO", "WOC", "WOS", "WO1", "NOACCESS"]:
                    dc = 1

                # Any unused bits on the right side of the LSB?
                while next_lsb < lsb:
                    mode[next_lsb] = "RO"
                    next_lsb += 1

                for _ in range(w):
                    mode[next_lsb] = field_access
                    dc_mask[next_lsb] = dc
                    next_lsb += 1
            # Any unused bits on the left side of the MSB?
            while next_lsb < UVM_REG_DATA_WIDTH:
                mode[next_lsb] = "RO"
                next_lsb += 1

            uvm_info("UVMRegBitBashSeq", sv.sformatf("Verifying bits in register %s in map \"%s\"...",
                rg.get_full_name(), maps[j].get_full_name()),UVM_LOW)

            bits_bashed = 0
            # Bash the kth bit
            for k in range(n_bits):
                # Cannot test unpredictable bit behavior
                if dc_mask[k] == 1:
                    continue
                dc_mask_int = 0x0
                for i in range(len(dc_mask)):
                    dc_mask_int |= (dc_mask[i] << i)

                await self.bash_kth_bit(rg, k, mode[k], maps[j], dc_mask_int)
                bits_bashed += 1

            if bits_bashed == 0:
                uvm_warning('UVMRegBitBashSeq', 'No bits bashed for ' +
                    rg.get_name())


    async def bash_kth_bit(self, rg, k, mode, _map, dc_mask):
        status = 0
        val = 0x0
        exp = 0x0
        v = 0x0
        bit_val = False
        uvm_info("UVMRegBitBashSeq", sv.sformatf("...Bashing %s bit #%0d", mode, k), UVM_HIGH)

        for _ in range(2):
            val = rg.get()
            v   = val
            exp = val
            mask = 1 << k
            #val[k] = ~val[k]
            #val = (val & ~mask) | (~val & mask)
            val ^= mask
            bit_val = (val >> k) & 0x1

            status = []
            await rg.write(status, val, UVM_FRONTDOOR, _map, self)
            status = status[0]
            if status != UVM_IS_OK:
                uvm_error("UVMRegBitBashSeq",
                        sv.sformatf("Status was %s when writing to register \"%s\" through map \"%s\".",
                            status.name(), rg.get_full_name(), _map.get_full_name()))

            exp = rg.get() & ~dc_mask
            status = []
            out_val = []
            await rg.read(status, out_val, UVM_FRONTDOOR, _map, self)
            status = status[0]
            val = out_val[0]
            if status != UVM_IS_OK:
               uvm_error("UVMRegBitBashSeq",
                       sv.sformatf("Status was %s when reading register \"%s\" through map \"%s\".",
                           status.name(), rg.get_full_name(), _map.get_full_name()))

            val &= ~dc_mask
            if val != exp:
                uvm_error("UVMRegBitBashSeq", sv.sformatf(
                    ("Writing a %b in bit #%0d of register \"%s\" with initial value "
                    + "'h%h yielded 'h%h instead of 'h%h"),
                    bit_val, k, rg.get_full_name(), v, val, exp))


uvm_object_utils(UVMRegSingleBitBashSeq)


#//------------------------------------------------------------------------------
#// Class: UVMRegBitBashSeq
#//
#//
#// Verify the implementation of all registers in a block
#// by executing the <UVMRegSingleBitBashSeq> sequence on it.
#//
#// If bit-type resource named
#// "NO_REG_TESTS" or "NO_REG_BIT_BASH_TEST"
#// in the "REG::" namespace
#// matches the full name of the block,
#// the block is not tested.
#//
#//| uvm_resource_db#(bit)::set({"REG::",regmodel.blk.get_full_name(),".*"},
#//|                            "NO_REG_TESTS", 1, this);
#//
#//------------------------------------------------------------------------------


class UVMRegBitBashSeq(UVMRegSequence):  # (uvm_sequence #(uvm_reg_item))

    def __init__(self, name="UVMRegBitBashSeq"):
        super().__init__(name)
        #   // Variable: reg_seq
        #   //
        #   // The sequence used to test one register
        #   //
        self.reg_seq = None
        #   // Variable: model
        #   //
        #   // The block to be tested. Declared in the base class.
        #   //
        self.model = None

    #   // Task: body
    #   //
    #   // Executes the Register Bit Bash sequence.
    #   // Do not call directly. Use seq.start() instead.
    #   //
    async def body(self):

        if self.model is None:
            uvm_error("UVMRegBitBashSeq", "No register model specified to run sequence on")
            return

        uvm_info("STARTING_SEQ","\n\nStarting " + self.get_name() + " sequence...\n",UVM_LOW)
        self.reg_seq = UVMRegSingleBitBashSeq.type_id.create("reg_single_bit_bash_seq")

        await self.reset_blk(self.model)
        self.model.reset()
        await self.do_block(self.model)


    #   // Task: do_block
    #   //
    #   // Test all of the registers in a given ~block~
    async def do_block(self, blk):
        regs = []

        if (UVMResourceDb.get_by_name("REG::" + blk.get_full_name(),
            "NO_REG_TESTS", 0) is not None or
            UVMResourceDb.get_by_name("REG::" + blk.get_full_name(),
                "NO_REG_BIT_BASH_TEST", 0) is not None):
            return

        # Iterate over all registers, checking accesses
        blk.get_registers(regs, UVM_NO_HIER)
        for i in range(len(regs)):
            # Registers with some attributes are not to be tested
            if (UVMResourceDb.get_by_name("REG::" + regs[i].get_full_name(),
                "NO_REG_TESTS", 0) is not None or
                UVMResourceDb.get_by_name("REG::" + regs[i].get_full_name(),
                    "NO_REG_BIT_BASH_TEST", 0) is not None):
                continue

            self.reg_seq.rg = regs[i]
            await self.reg_seq.start(None,self)

        blks = []
        blk.get_blocks(blks)
        for blk in blks:
            await self.do_block(blk)


    #   // Task: reset_blk
    #   //
    #   // Reset the DUT that corresponds to the specified block abstraction class.
    #   //
    #   // Currently empty.
    #   // Will rollback the environment's phase to the ~reset~
    #   // phase once the new phasing is available.
    #   //
    #   // In the meantime, the DUT should be reset before executing this
    #   // test sequence or this method should be implemented
    #   // in an extension to reset the DUT.
    #   //
    async def reset_blk(self, blk):
        await uvm_zero_delay()


uvm_object_utils(UVMRegBitBashSeq)
